# /// script
# dependencies = [
#     "ndv[pyqt,vispy]",
#     "torch>=2.5",
# ]
# ///
from __future__ import annotations

try:
    import torch
except ImportError:
    raise ImportError("Please install torch to run this example")

import warnings

import ndv

warnings.filterwarnings("ignore", "Named tensors")  # Named tensors are experimental

# Example usage
try:
    torch_data = torch.tensor(  # type: ignore [call-arg]
        ndv.data.nd_sine_wave(),
        names=("t", "c", "z", "y", "x"),
    )
except TypeError:
    print("Named tensors are not supported in your version of PyTorch")
    torch_data = torch.tensor(ndv.data.nd_sine_wave())

ndv.imshow(torch_data, visible_axes=("y", -1))



# /// script
# dependencies = [
#     "ndv[pyqt,pygfx]",
#     "imageio[tifffile]",
# ]
# ///
"""Example of streaming data."""

import numpy as np

import ndv

# some data we're going to stream (as if it was coming from a camera)
data = ndv.data.cells3d()[:, 1]

# a buffer to hold the current frame in the viewer
buffer = np.zeros_like(data[0])
viewer = ndv.ArrayViewer(buffer)
viewer.show()


# function that will be called after the app is running
def stream(nframes: int = len(data) * 4) -> None:
    for i in range(nframes):
        # iterate over the data, update the buffer *in place*,
        buffer[:] = data[i % len(data)]
        # and update the viewer index to redraw
        viewer.display_model.current_index.update()
        ndv.process_events()  # force viewer updates for this example


ndv.call_later(200, stream)
ndv.run_app()



"""Tests pertaining to VisPy components"""

from importlib.util import find_spec

import pytest

if not find_spec("vispy"):
    pytest.skip(
        "Skipping vispy tests as vispy is not installed", allow_module_level=True
    )



from __future__ import annotations

import contextlib
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import cmap
import ipywidgets as widgets
import psygnal
from IPython.display import Javascript, display

from ndv.models._array_display_model import ChannelMode
from ndv.models._lut_model import ClimPolicy, ClimsManual, ClimsPercentile
from ndv.models._viewer_model import ArrayViewerModel, InteractionMode
from ndv.views.bases import ArrayView, LUTView

if TYPE_CHECKING:
    from collections.abc import Container, Hashable, Iterator, Mapping, Sequence

    from psygnal import EmissionInfo
    from traitlets import HasTraits
    from vispy.app.backends import _jupyter_rfb

    from ndv._types import AxisKey, ChannelKey
    from ndv.views.bases._graphics._canvas import HistogramCanvas

# not entirely sure why it's necessary to specifically annotat signals as : PSignal
# i think it has to do with type variance?


@contextlib.contextmanager
def notifications_blocked(
    obj: HasTraits, name: str = "value", type: str = "change"
) -> Iterator[None]:
    # traitlets doesn't provide a public API for this
    notifiers: list | None = obj._trait_notifiers.get(name, {}).pop(type, None)
    try:
        yield
    finally:
        if notifiers is not None:
            obj._trait_notifiers[name][type] = notifiers


class RightClickButton(widgets.ToggleButton):
    """Custom Button widget that shows a popup on right-click."""

    # TODO: These are likely unnecessary
    # _right_click_triggered = Bool(False).tag(sync=True)
    # popup_content = Unicode("Right-click menu").tag(sync=True)

    def __init__(self, channel: ChannelKey, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._channel = channel
        self.add_class(f"right-click-button{channel}")
        self.add_right_click_handler()

        self.lower_tail = widgets.BoundedFloatText(
            value=0.0,
            min=0.0,
            max=100.0,
            step=0.1,
            description="Ignore Lower Tail:",
            style={"description_width": "initial"},
        )
        self.upper_tail = widgets.BoundedFloatText(
            value=0.0,
            min=0.0,
            max=100.0,
            step=0.1,
            description="Ignore Upper Tail:",
            style={"description_width": "initial"},
        )
        self.popup_content = widgets.VBox(
            [self.lower_tail, self.upper_tail],
            layout=widgets.Layout(
                display="none",
            ),
        )
        self.popup_content.add_class(f"ipywidget-popup{channel}")
        display(self.popup_content)  # type: ignore [no-untyped-call]

    def add_right_click_handler(self) -> None:
        # fmt: off
        js_code = ( """
        (function() {
            function setup_rightclick() {
                // Get all buttons with the right-click-button class
                var button = document.getElementsByClassName(
                    'right-click-button""" + f"{self._channel}" + """'
                )[0];
                if (!button) {
                    return;
                }

                // For each button, add a contextmenu listener
                button.addEventListener('contextmenu', function(e) {
                    // Prevent default context menu
                    e.preventDefault();
                    e.stopPropagation();

                    // Get button position
                    var rect = this.getBoundingClientRect();
                    var scrollLeft = window.pageXOffset ||
                        document.documentElement.scrollLeft;
                    var scrollTop = window.pageYOffset ||
                        document.documentElement.scrollTop;

                    // Position the popup above the button
                    var popup = document.getElementsByClassName(
                        'ipywidget-popup""" + f"{self._channel}" + """'
                    )[0];
                    popup.style.display = '';
                    popup.style.position = 'absolute';
                    popup.style.top = (rect.bottom + scrollTop) + 'px';
                    popup.style.left = (rect.left + scrollLeft) + 'px';

                    // Style the popup
                    popup.style.background = 'white';
                    popup.style.border = '1px solid #ccc';
                    popup.style.borderRadius = '3px';
                    popup.style.padding = '8px';
                    popup.style.boxShadow = '0 2px 5px rgba(0,0,0,0.2)';
                    popup.style.zIndex = '1000';

                    // Add to body
                    document.body.appendChild(popup);

                    // Close popup when clicking elsewhere
                    document.addEventListener('click', function closePopup(event) {
                        var popup = document.getElementsByClassName(
                            'ipywidget-popup""" + f"{self._channel}" + """'
                        )[0];
                        if (popup && !popup.contains(event.target)) {
                            popup.style.display = 'none';
                            document.removeEventListener('click', closePopup);
                        }
                    });

                    return false;
                });
            }

            // Make sure it works even after widget is redrawn/updated
            setTimeout(setup_rightclick, 1000);
        })();
        """)
        # fmt: on
        display(Javascript(js_code))  # type: ignore [no-untyped-call]


class JupyterLUTView(LUTView):
    # NB: In practice this will be a ChannelKey but Unions not allowed here.
    histogramRequested = psygnal.Signal(object)

    def __init__(
        self,
        channel: ChannelKey = None,
        default_luts: Sequence[Any] = ("gray", "green", "magenta", "red", "blue"),
    ) -> None:
        self._channel = channel
        self._histogram: HistogramCanvas | None = None
        # WIDGETS
        self._visible = widgets.Checkbox(value=True, indent=False)
        self._visible.layout.width = "60px"
        _luts = [cmap.Colormap(x).name.split(":")[-1] for x in default_luts]
        self._cmap = widgets.Dropdown(options=_luts, value=_luts[0])
        self._cmap.layout.width = "200px"
        self._clims = widgets.FloatRangeSlider(
            value=[0, 2**16],
            min=0,
            max=2**16,
            step=1,
            orientation="horizontal",
            readout=True,
            readout_format=".0f",
        )
        self._clims.layout.width = "100%"
        self._auto_clim = RightClickButton(
            channel=channel,
            value=True,
            description="Auto",
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Auto scale",
            layout=widgets.Layout(min_width="40px"),
        )
        self._histogram_btn = widgets.ToggleButton(
            value=False,
            description="",
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            icon="bar-chart",
            tooltip="View Histogram",
            layout=widgets.Layout(width="40px"),
        )

        histogram_ctrl_width = 40
        self._log = widgets.ToggleButton(
            value=False,
            description="log",
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Apply logarithm (base 10, count+1) to bins",
            layout=widgets.Layout(width=f"{histogram_ctrl_width}px"),
        )
        self._reset_histogram = widgets.Button(
            description="",
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            icon="expand",
            tooltip="Reset histogram view",
            layout=widgets.Layout(width=f"{histogram_ctrl_width}px"),
        )

        # LAYOUT

        lut_ctrls = widgets.HBox(
            [
                self._visible,
                self._cmap,
                self._clims,
                self._auto_clim,
                self._histogram_btn,
            ]
        )

        histogram_ctrls = widgets.VBox(
            [self._log, self._reset_histogram],
            layout=widgets.Layout(
                # Avoids scrollbar on buttons
                min_width=f"{histogram_ctrl_width + 10}px",
                # Floats buttons to the bottom
                justify_content="flex-end",
            ),
        )

        self._histogram_container = widgets.HBox(
            # Note that we'll add a histogram here later
            [histogram_ctrls],
            layout=widgets.Layout(
                # Constrains histogram to 100px tall
                max_height="100px",
                # Avoids vertical scrollbar from
                # histogram being *just a bit* too tall
                overflow="hidden",
                # Hide histogram initially
                display="none",
            ),
        )
        self.layout = widgets.VBox([lut_ctrls, self._histogram_container])

        # CONNECTIONS
        self._visible.observe(self._on_visible_changed, names="value")
        self._cmap.observe(self._on_cmap_changed, names="value")
        self._clims.observe(self._on_clims_changed, names="value")
        self._auto_clim.observe(self._on_autoscale_changed, names="value")
        self._auto_clim.lower_tail.observe(self._on_auto_tails_changed, names="value")
        self._auto_clim.upper_tail.observe(self._on_auto_tails_changed, names="value")
        self._histogram_btn.observe(self._on_histogram_requested, names="value")
        self._log.observe(self._on_log_toggled, names="value")
        self._reset_histogram.on_click(self._on_reset_histogram_clicked)

    # ------------------ emit changes to the controller ------------------

    def _on_clims_changed(self, change: dict[str, Any]) -> None:
        if self._model:
            clims = self._clims.value
            self._model.clims = ClimsManual(min=clims[0], max=clims[1])

    def _on_visible_changed(self, change: dict[str, Any]) -> None:
        if self._model:
            self._model.visible = self._visible.value

    def _on_cmap_changed(self, change: dict[str, Any]) -> None:
        if self._model:
            self._model.cmap = self._cmap.value

    def _on_autoscale_changed(self, change: dict[str, Any]) -> None:
        if self._model:
            if change["new"]:  # Autoscale
                lower_tail = self._auto_clim.lower_tail.value
                upper_tail = self._auto_clim.upper_tail.value
                self._model.clims = ClimsPercentile(
                    min_percentile=lower_tail, max_percentile=100 - upper_tail
                )
            else:  # Manually scale
                clims = self._clims.value
                self._model.clims = ClimsManual(min=clims[0], max=clims[1])

    def _on_auto_tails_changed(self, change: dict[str, Any]) -> None:
        # Update clim policy if autoscaling is active
        if self._auto_clim.value:
            self._on_autoscale_changed({"new": True})

    def _on_histogram_requested(self, change: dict[str, Any]) -> None:
        # Generate the histogram if we haven't done so yet
        if not self._histogram:
            self.histogramRequested.emit(self._channel)
        # show or hide the histogram controls
        self._histogram_container.layout.display = "flex" if change["new"] else "none"

    def _on_log_toggled(self, change: dict[str, Any]) -> None:
        if hist := self._histogram:
            hist.set_log_base(10 if change["new"] else None)

    def _on_reset_histogram_clicked(self, change: dict[str, Any]) -> None:
        self._log.value = False
        if hist := self._histogram:
            hist.set_range()

    # ------------------ receive changes from the controller ---------------

    def set_channel_name(self, name: str) -> None:
        self._visible.description = name

    def set_clim_policy(self, policy: ClimPolicy) -> None:
        with notifications_blocked(self._auto_clim):
            self._auto_clim.value = not policy.is_manual
            if isinstance(policy, ClimsPercentile):
                self._auto_clim.lower_tail.value = policy.min_percentile
                self._auto_clim.upper_tail.value = 100 - policy.max_percentile

    def set_colormap(self, cmap: cmap.Colormap) -> None:
        name = cmap.name.split(":")[-1]
        if name not in self._cmap.options:
            self._cmap.options = (*self._cmap.options, name)
        self._cmap.value = name

    def set_clims(self, clims: tuple[float, float]) -> None:
        # block self._clims.observe, otherwise autoscale will be forced off
        with notifications_blocked(self._clims):
            # FIXME: Internally the clims are being rounded to whole numbers.
            # The rounding is somehow avoiding notifications_blocked.
            self._clims.value = [int(c) for c in clims]

    def set_clim_bounds(
        self,
        bounds: tuple[float | None, float | None] = (None, None),
    ) -> None:
        mi = 0 if bounds[0] is None else int(bounds[0])
        ma = 65535 if bounds[1] is None else int(bounds[1])
        # block self._clims.observe, otherwise autoscale will be forced off
        with notifications_blocked(self._clims):
            self._clims.min = mi
            self._clims.max = ma
            current_value = self._clims.value
            self._clims.value = (max(current_value[0], mi), min(current_value[1], ma))

    def set_channel_visible(self, visible: bool) -> None:
        self._visible.value = visible

    def set_gamma(self, gamma: float) -> None:
        pass

    def set_visible(self, visible: bool) -> None:
        # show or hide the actual widget itself
        self.layout.layout.display = "flex" if visible else "none"

    def close(self) -> None:
        self.layout.close()

    def frontend_widget(self) -> Any:
        return self.layout

    # ------------------ private methods ---------------

    def add_histogram(self, histogram: HistogramCanvas) -> None:
        widget = histogram.frontend_widget()
        # Resize widget to a respectable size
        widget.set_trait("css_height", "auto")
        # Add widget to view
        self._histogram_container.children = (
            *self._histogram_container.children,
            widget,
        )
        self._histogram = histogram


class JupyterRGBView(JupyterLUTView):
    def __init__(self, channel: ChannelKey = None) -> None:
        super().__init__(channel)
        self._cmap.layout.display = "none"


SPIN_GIF = str(Path(__file__).parent.parent / "_resources" / "spin.gif")


class JupyterArrayView(ArrayView):
    def __init__(
        self,
        canvas_widget: _jupyter_rfb.CanvasBackend,
        viewer_model: ArrayViewerModel,
    ) -> None:
        self._viewer_model = viewer_model
        self._viewer_model.events.connect(self._on_viewer_model_event)
        # WIDGETS
        self._canvas_widget = canvas_widget
        self._visible_axes: Sequence[AxisKey] = []
        self._luts: dict[ChannelKey, JupyterLUTView] = {}

        self._sliders: dict[Hashable, widgets.IntSlider] = {}
        self._slider_box = widgets.VBox([], layout=widgets.Layout(width="100%"))
        self._luts_box = widgets.VBox([], layout=widgets.Layout(width="100%"))

        # labels for data and hover info
        self._data_info_label = widgets.Label()
        self._hover_info_label = widgets.Label()

        # spinner to indicate progress
        self._progress_spinner = widgets.Image.from_file(
            SPIN_GIF, width=18, height=18, layout=widgets.Layout(display="none")
        )

        # the button that controls the display mode of the channels
        self._channel_mode_combo = widgets.Dropdown(
            options=[ChannelMode.GRAYSCALE, ChannelMode.COMPOSITE, ChannelMode.RGBA],
            value=str(ChannelMode.GRAYSCALE),
        )
        self._channel_mode_combo.layout.width = "120px"
        self._channel_mode_combo.layout.align_self = "flex-end"
        self._channel_mode_combo.observe(self._on_channel_mode_changed, names="value")

        # Reset zoom button
        self._reset_zoom_btn = widgets.Button(
            tooltip="Reset Zoom",
            icon="expand",
            layout=widgets.Layout(width="40px"),
        )
        self._reset_zoom_btn.on_click(self._on_reset_zoom_clicked)

        # 3d view button
        self._ndims_btn = widgets.ToggleButton(
            value=False,
            description="3D",
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="View in 3D",
            layout=widgets.Layout(width="40px"),
        )
        self._ndims_btn.observe(self._on_ndims_toggled, names="value")

        # Add ROI button
        self._add_roi_btn = widgets.ToggleButton(
            value=False,
            description="New ROI",
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Adds a new Rectangular ROI.",
            icon="square",
        )

        self._add_roi_btn.observe(self._on_add_roi_button_toggle, names="value")

        # LAYOUT

        top_row = widgets.HBox(
            [self._data_info_label, self._progress_spinner],
            layout=widgets.Layout(
                display="flex",
                justify_content="space-between",
                align_items="center",
            ),
        )

        try:
            width = getattr(canvas_widget, "css_width", "600px").replace("px", "")
            width = f"{int(width) + 4}px"
        except Exception:
            width = "604px"

        self._btns_box = widgets.HBox(
            [
                self._channel_mode_combo,
                self._ndims_btn,
                self._add_roi_btn,
                self._reset_zoom_btn,
            ],
            layout=widgets.Layout(justify_content="flex-end"),
        )
        self.layout = widgets.VBox(
            [
                top_row,
                self._canvas_widget,
                self._hover_info_label,
                self._slider_box,
                self._luts_box,
                self._btns_box,
            ],
            layout=widgets.Layout(width=width),
        )

        # CONNECTIONS

        self._channel_mode_combo.observe(self._on_channel_mode_changed, names="value")

    def create_sliders(self, coords: Mapping[Hashable, Sequence]) -> None:
        """Update sliders with the given coordinate ranges."""
        sliders = []
        self._sliders.clear()
        for axis, _coords in coords.items():
            if not isinstance(_coords, range):
                raise NotImplementedError("Only range is supported for now")

            sld = widgets.IntSlider(
                value=_coords.start,
                min=_coords.start,
                max=_coords.stop - 1,
                step=_coords.step,
                description=str(axis),
                continuous_update=True,
                orientation="horizontal",
            )
            sld.layout.width = "99%"
            sld.observe(self._on_slider_change, "value")
            sliders.append(sld)
            self._sliders[axis] = sld
        self._slider_box.children = sliders

        self.currentIndexChanged.emit()

    def hide_sliders(
        self, axes_to_hide: Container[Hashable], show_remainder: bool = True
    ) -> None:
        """Hide sliders based on visible axes."""
        for ax, slider in self._sliders.items():
            if ax in axes_to_hide:
                slider.layout.display = "none"
            elif show_remainder:
                slider.layout.display = "flex"

    def current_index(self) -> Mapping[AxisKey, int | slice]:
        """Return the current value of the sliders."""
        return {axis: slider.value for axis, slider in self._sliders.items()}

    def set_current_index(self, value: Mapping[AxisKey, int | slice]) -> None:
        """Set the current value of the sliders."""
        changed = False
        # this type ignore is only necessary because we had to override the signal
        # to be a PSignal in the class def above :(
        with self.currentIndexChanged.blocked():
            for axis, val in value.items():
                if isinstance(val, slice):
                    raise NotImplementedError("Slices are not supported yet")

                if sld := self._sliders.get(axis):
                    if sld.value != val:
                        sld.value = val
                        changed = True
                else:  # pragma: no cover
                    warnings.warn(f"Axis {axis} not found in sliders", stacklevel=2)
        if changed:
            self.currentIndexChanged.emit()

    def add_lut_view(self, channel: ChannelKey) -> JupyterLUTView:
        """Add a LUT view to the viewer."""
        wdg = (
            JupyterRGBView(channel)
            if channel == "RGB"
            else JupyterLUTView(channel, self._viewer_model.default_luts)
        )
        layout = self._luts_box
        self._luts[channel] = wdg

        wdg.histogramRequested.connect(self.histogramRequested)
        layout.children = (*layout.children, wdg.layout)
        return wdg

    def remove_lut_view(self, view: LUTView) -> None:
        """Remove a LUT view from the viewer."""
        if view not in self._luts.values():
            return
        view = cast("JupyterLUTView", view)
        layout = self._luts_box
        layout.children = tuple(
            wdg for wdg in layout.children if wdg != view.frontend_widget()
        )

    def set_data_info(self, data_info: str) -> None:
        self._data_info_label.value = data_info

    def set_hover_info(self, hover_info: str) -> None:
        self._hover_info_label.value = hover_info

    def set_channel_mode(self, mode: ChannelMode) -> None:
        with self.channelModeChanged.blocked():
            self._channel_mode_combo.value = mode.value

    def _on_slider_change(self, change: dict[str, Any]) -> None:
        """Emit signal when a slider value changes."""
        self.currentIndexChanged.emit()

    def _on_channel_mode_changed(self, change: dict[str, Any]) -> None:
        """Emit signal when the channel mode changes."""
        self.channelModeChanged.emit(ChannelMode(change["new"]))

    def add_histogram(self, channel: ChannelKey, histogram: HistogramCanvas) -> None:
        if lut := self._luts.get(channel, None):
            lut.add_histogram(histogram)

    def _on_add_roi_button_toggle(self, change: dict[str, Any]) -> None:
        """Emit signal when the channel mode changes."""
        self._viewer_model.interaction_mode = (
            InteractionMode.CREATE_ROI if change["new"] else InteractionMode.PAN_ZOOM
        )

    def remove_histogram(self, widget: Any) -> None:
        """Remove a histogram widget from the viewer."""

    def frontend_widget(self) -> Any:
        return self.layout

    def set_visible(self, visible: bool) -> None:
        # show or hide the actual widget itself
        from IPython import display

        if visible:
            display.display(self.layout)  # type: ignore [no-untyped-call]
        else:
            display.clear_output()  # type: ignore [no-untyped-call]

    def visible_axes(self) -> Sequence[AxisKey]:
        return self._visible_axes

    def set_visible_axes(self, axes: Sequence[AxisKey]) -> None:
        self._visible_axes = tuple(axes)
        with notifications_blocked(self._ndims_btn):
            self._ndims_btn.value = len(axes) > 2

    def _on_ndims_toggled(self, change: dict[str, Any]) -> None:
        self.ndimToggleRequested.emit(change["new"])

    def _on_reset_zoom_clicked(self, change: dict[str, Any]) -> None:
        self.resetZoomClicked.emit()

    def close(self) -> None:
        self.layout.close()

    def _on_viewer_model_event(self, info: EmissionInfo) -> None:
        sig_name = info.signal.name
        value = info.args[0]
        if sig_name == "show_progress_spinner":
            self._progress_spinner.layout.display = "flex" if value else "none"
        elif sig_name == "interaction_mode":
            # If leaving CanvasMode.CREATE_ROI, uncheck the ROI button
            _new, old = info.args
            if old == InteractionMode.CREATE_ROI:
                self._add_roi_btn.value = False
        elif sig_name == "show_histogram_button":
            # Note that "block" displays the icon better than "flex"
            for lut in self._luts.values():
                lut._histogram_btn.layout.display = "block" if value else "none"
        elif sig_name == "show_roi_button":
            self._add_roi_btn.layout.display = "flex" if value else "none"
        elif sig_name == "show_channel_mode_selector":
            self._channel_mode_combo.layout.display = "flex" if value else "none"
        elif sig_name == "show_reset_zoom_button":
            self._reset_zoom_btn.layout.display = "flex" if value else "none"
        elif sig_name == "show_3d_button":
            self._ndims_btn.layout.display = "flex" if value else "none"
        # elif sig_name == "show_play_button":
        #     ...
        elif sig_name == "show_data_info":
            self._data_info_label.display = "flex" if value else None
            self._hover_info_label.display = "flex" if value else None
        elif sig_name == "show_controls":
            # Show or hide the entire controls area (dims sliders + LUTs + btns)
            self._slider_box.display = "flex" if value else None
            self._btns_box.display = "flex" if value else None
            self._luts_box.display = "flex" if value else None



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import MagicMock

import cmap
from pytest import fixture
from qtpy.QtWidgets import QWidget

from ndv.models._lut_model import ClimsManual, ClimsMinMax, ClimsPercentile, LUTModel
from ndv.views._qt._array_view import QLUTView
from ndv.views.bases._graphics._canvas import HistogramCanvas

if TYPE_CHECKING:
    from pytestqt.qtbot import QtBot


@fixture
def model() -> LUTModel:
    return LUTModel()


@fixture
def view(model: LUTModel, qtbot: QtBot) -> QLUTView:
    view = QLUTView()
    qtbot.add_widget(view.frontend_widget())
    # Set the model
    assert view.model is None
    view.model = model
    assert view.model is model
    return view


def test_QLUTView_update_model(model: LUTModel, view: QLUTView) -> None:
    """Ensures the view updates when the model is changed."""

    # Test modifying model.clims
    assert view._qwidget.auto_clim.isChecked()
    model.clims = ClimsManual(min=0, max=1)
    assert not view._qwidget.auto_clim.isChecked()
    model.clims = ClimsPercentile(min_percentile=0, max_percentile=100)
    assert view._qwidget.auto_clim.isChecked()
    model.clims = ClimsPercentile(min_percentile=1, max_percentile=99)
    assert view._qwidget.lower_tail.value() == 1
    assert view._qwidget.upper_tail.value() == 1

    # Test modifying model.visible
    assert view._qwidget.visible.isChecked()
    model.visible = False
    assert not view._qwidget.visible.isChecked()
    model.visible = True
    assert view._qwidget.visible.isChecked()

    # Test modifying model.cmap
    new_cmap = cmap.Colormap("red")
    assert view._qwidget.cmap.currentColormap() != new_cmap
    model.cmap = new_cmap
    assert view._qwidget.cmap.currentColormap() == new_cmap


def test_QLUTView_update_view(model: LUTModel, view: QLUTView) -> None:
    """Ensures the model updates when the view is changed."""

    new_visible = not model.visible
    view._qwidget.visible.setChecked(new_visible)
    assert view._qwidget.visible.isChecked() == new_visible

    new_cmap = view._qwidget.cmap.itemColormap(1)
    assert model.cmap != new_cmap
    assert new_cmap is not None
    view._qwidget.cmap.setCurrentIndex(1)
    assert model.cmap == new_cmap

    # Test toggling auto_clim
    assert model.clims == ClimsPercentile(min_percentile=0, max_percentile=100)
    view._qwidget.auto_clim.setChecked(False)
    mi, ma = view._qwidget.clims.value()
    assert model.clims == ClimsManual(min=mi, max=ma)
    view._qwidget.auto_clim.setChecked(True)
    assert model.clims == ClimsPercentile(min_percentile=0, max_percentile=100)

    # Test modifying tails changes percentiles
    view._qwidget.lower_tail.setValue(0.1)
    assert model.clims == ClimsPercentile(min_percentile=0.1, max_percentile=100)
    view._qwidget.upper_tail.setValue(0.2)
    assert model.clims == ClimsPercentile(min_percentile=0.1, max_percentile=99.8)

    # When gui clims change, autoscale should be disabled
    model.clims = ClimsMinMax()
    view._qwidget.clims.setValue((0, 1))
    assert model.clims == ClimsManual(min=0, max=1)


def test_QLUTView_histogram_controls(model: LUTModel, view: QLUTView) -> None:
    # Mock up a histogram
    hist_mock = MagicMock(spec=HistogramCanvas)
    hist_frontend = QWidget()
    hist_mock.frontend_widget.return_value = hist_frontend

    # Add the histogram and assert it was correctly added
    view._add_histogram(hist_mock)
    assert view.histogram is hist_mock

    # Assert histogram button toggles visibility
    # Note that the parent must be visible for child visibility to change
    old_visibility = view._qwidget.histogram_btn.isChecked()
    view._qwidget.setVisible(True)
    assert not view._qwidget.histogram_btn.isChecked()

    view._qwidget.histogram_btn.setChecked(True)
    assert view._qwidget.hist_log.isVisible()
    assert view._qwidget.hist_range.isVisible()
    assert hist_frontend.isVisible()

    view._qwidget.histogram_btn.setChecked(False)
    assert not view._qwidget.hist_log.isVisible()
    assert not view._qwidget.hist_range.isVisible()
    assert not hist_frontend.isVisible()
    view._qwidget.setVisible(old_visibility)

    # Assert toggling the log button alters the logarithmic base
    view._qwidget.hist_log.setChecked(True)
    hist_mock.set_log_base.assert_called_once_with(10)
    hist_mock.reset_mock()
    view._qwidget.hist_log.setChecked(False)
    hist_mock.set_log_base.assert_called_once_with(None)
    hist_mock.reset_mock()

    # Assert pressing the reset view button sets the histogram range
    view._qwidget.hist_range.click()
    hist_mock.set_range.assert_called_once_with()
    hist_mock.reset_mock()

    # Assert pressing the reset view button turns off log mode
    view._qwidget.hist_log.setChecked(True)
    view._qwidget.hist_range.click()
    assert not view._qwidget.hist_log.isChecked()
    hist_mock.reset_mock()



from __future__ import annotations

import gc
import importlib
import importlib.util
import os
import sys
from collections.abc import Iterator
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any
from unittest.mock import patch

import pytest

from ndv.views import gui_frontend
from ndv.views._app import GUI_ENV_VAR, GuiFrontend

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop
    from collections.abc import Iterator

    import wx
    from pytest import FixtureRequest
    from qtpy.QtWidgets import QApplication


@pytest.fixture
def asyncio_app() -> Iterator[AbstractEventLoop]:
    import asyncio

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    yield loop
    loop.close()


@pytest.fixture(scope="session")
def wxapp() -> Iterator[wx.App]:
    import wx

    # wx.App() hangs on macOS if a QApplication is already running, because both
    # try to own the NSApplication singleton. Skip wx tests in that case.
    for mod_name in ("PyQt5", "PySide2", "PySide6", "PyQt6"):
        if mod := sys.modules.get(f"{mod_name}.QtWidgets"):
            if (qapp := getattr(mod, "QApplication", None)) and qapp.instance():
                pytest.skip("Qt already running. Cannot also test wx.")

    if (_wxapp := wx.App.Get()) is None:
        _wxapp = wx.App()
    yield _wxapp


@pytest.fixture
def any_app(request: pytest.FixtureRequest) -> Iterator[Any]:
    # this fixture will use the appropriate application depending on the env var
    # NDV_GUI_FRONTEND='qt' pytest
    # NDV_GUI_FRONTEND='jupyter' pytest
    try:
        frontend = gui_frontend()
    except RuntimeError:
        # if we don't find any frontend, and jupyter is available, use that
        # since it requires very little setup
        if importlib.util.find_spec("jupyter"):
            os.environ[GUI_ENV_VAR] = "jupyter"

        frontend = gui_frontend()

    if frontend == GuiFrontend.QT:
        app = request.getfixturevalue("qapp")
        qtbot = request.getfixturevalue("qtbot")
        with patch.object(app, "exec", lambda *_: app.processEvents()):
            with _catch_qt_leaks(request, app):
                yield app, qtbot
    elif frontend == GuiFrontend.JUPYTER:
        yield request.getfixturevalue("asyncio_app")
    elif frontend == GuiFrontend.WX:
        yield request.getfixturevalue("wxapp")
    else:
        raise RuntimeError("No GUI frontend found")


@contextmanager
def _catch_qt_leaks(request: FixtureRequest, qapp: QApplication) -> Iterator[None]:
    """Run after each test to ensure no widgets have been left around.

    When this test fails, it means that a widget being tested has an issue closing
    cleanly. Perhaps a strong reference has leaked somewhere.  Look for
    `functools.partial(self._method)` or `lambda: self._method` being used in that
    widget's code.
    """
    # check for the "allow_leaks" marker
    if "allow_leaks" in request.node.keywords:
        yield
        return

    nbefore = len(qapp.topLevelWidgets())
    failures_before = request.session.testsfailed
    yield
    # if the test failed, don't worry about checking widgets
    if request.session.testsfailed - failures_before:
        return
    allow: list[type] = []
    try:
        from vispy.app.backends._qt import CanvasBackendDesktop

        allow.append(CanvasBackendDesktop)
    except (ImportError, RuntimeError):
        pass
    try:
        # This is a known widget that is not cleaned up properly
        # but it's not clear how to fix it
        from rendercanvas.qt import QRenderWidget

        allow.append(QRenderWidget)
    except (ImportError, RuntimeError):
        pass

    # This is a known widget that is not cleaned up properly
    remaining = [w for w in qapp.topLevelWidgets() if not isinstance(w, tuple(allow))]
    if len(remaining) > nbefore:
        test_node = request.node

        test = f"{test_node.path.name}::{test_node.originalname}"
        msg = f"{len(remaining)} topLevelWidgets remaining after {test!r}:"

        for widget in remaining:
            try:
                obj_name = widget.objectName()
            except Exception:
                obj_name = None
            msg += f"\n{widget!r} {obj_name!r}"
            # Get the referrers of the widget
            referrers = gc.get_referrers(widget)
            msg += "\n  Referrers:"
            for ref in referrers:
                if ref is remaining:
                    continue
                msg += f"\n  -   {ref}, {id(ref):#x}"

        raise AssertionError(msg)



from collections.abc import Callable, Sequence
from typing import Any, Protocol, SupportsIndex, TypeAlias, cast

import numpy as np
import numpy.typing as npt
from pydantic_core import core_schema

_ShapeLike: TypeAlias = SupportsIndex | Sequence[SupportsIndex]


class Reducer(Protocol):
    """Function to reduce an array along an axis.

    A reducer is any function that takes an array-like, and an optional axis argument,
    and returns a reduced array.  Examples include `np.max`, `np.mean`, etc.
    """

    def __call__(self, a: npt.ArrayLike, axis: _ShapeLike = ...) -> npt.ArrayLike:
        """Reduce an array along an axis."""
        raise NotImplementedError()


def _str_to_callable(obj: Any) -> Callable:
    """Deserialize a callable from a string."""
    if isinstance(obj, str):
        # e.g. "numpy.max" -> numpy.max
        try:
            mod_name, qual_name = obj.rsplit(".", 1)
            mod = __import__(mod_name, fromlist=[qual_name])
            obj = getattr(mod, qual_name)
        except Exception:
            try:
                # fallback to numpy
                # e.g. "max" -> numpy.max
                obj = getattr(np, obj)
            except Exception:
                raise

    if not callable(obj):
        raise TypeError(f"Expected a callable or string, got {type(obj)}")
    return cast("Callable", obj)


def _callable_to_str(obj: str | Callable) -> str:
    """Serialize a callable to a string."""
    if isinstance(obj, str):
        return obj
    # e.g. numpy.max -> "numpy.max"
    return f"{obj.__module__}.{obj.__qualname__}"


class ReducerType(Reducer):
    """Reducer type for pydantic.

    This just provides a pydantic core schema for a generic callable that accepts an
    array and an axis argument and returns an array (of reduced dimensionality).
    This serializes/deserializes the callable as a string (module.qualname).
    """

    @classmethod
    def __get_pydantic_core_schema__(cls, source: Any, handler: Any) -> Any:
        """Get the Pydantic schema for this object."""
        ser_schema = core_schema.plain_serializer_function_ser_schema(_callable_to_str)
        return core_schema.no_info_before_validator_function(
            _str_to_callable,
            # using callable_schema() would be more correct, but prevents dumping schema
            core_schema.any_schema(),
            serialization=ser_schema,
        )



from __future__ import annotations

import numpy as np
import pytest

from ndv.controllers._image_stats import (
    _compute_histogram,
    _compute_int_histogram,
    _minmax_from_counts,
    _percentile_from_counts,
    _stddev_from_counts,
    compute_image_stats,
)
from ndv.models._lut_model import ClimsManual, ClimsMinMax, ClimsPercentile, ClimsStdDev

RNG = np.random.default_rng(42)


@pytest.mark.parametrize(
    "data, bits, expected_nbins, check_indices",
    [
        (np.array([[0, 1, 2, 255]], dtype=np.uint8), 8, 256, [0, 255]),
        (np.array([[0, 100, 4095]], dtype=np.uint16), 12, 4096, [0, 100, 4095]),
    ],
)
def test_compute_int_histogram(
    data: np.ndarray, bits: int, expected_nbins: int, check_indices: list[int]
) -> None:
    counts, edges = _compute_int_histogram(data, bits=bits)
    assert counts.shape == (expected_nbins,)
    assert edges.shape == (expected_nbins + 1,)
    for idx in check_indices:
        assert counts[idx] == 1


@pytest.mark.parametrize(
    "counts, edges, expected",
    [
        (
            np.array([0, 0, 5, 3, 0, 2, 0]),
            np.arange(8, dtype=np.float64) - 0.5,
            (2.0, 5.0),
        ),
        (np.zeros(10, dtype=np.int64), np.arange(11, dtype=np.float64), (0.0, 0.0)),
    ],
)
def test_minmax_from_counts(
    counts: np.ndarray, edges: np.ndarray, expected: tuple[float, float]
) -> None:
    assert _minmax_from_counts(counts, edges) == pytest.approx(expected)


def test_percentile_from_counts() -> None:
    counts = np.zeros(256, dtype=np.int64)
    counts[10] = 50
    counts[200] = 50
    edges = np.arange(257, dtype=np.float64) - 0.5
    assert _percentile_from_counts(counts, edges, 25, 75) == pytest.approx(
        (10.0, 200.0)
    )


def test_stddev_from_counts() -> None:
    counts = np.zeros(256, dtype=np.int64)
    counts[128] = 1000
    edges = np.arange(257, dtype=np.float64) - 0.5
    # All data in one bin => std ≈ 0, so lo ≈ hi ≈ center
    assert _stddev_from_counts(counts, edges, n_stdev=2, center=None) == pytest.approx(
        (128.0, 128.0)
    )


@pytest.mark.parametrize("need_histogram", [False, True])
def test_manual_clims(need_histogram: bool) -> None:
    data = RNG.integers(0, 255, (100, 100), dtype=np.uint8)
    stats = compute_image_stats(
        data, ClimsManual(min=10, max=200), need_histogram=need_histogram
    )
    assert stats.clims == (10, 200)
    assert (stats.counts is not None) == need_histogram
    assert (stats.bin_edges is not None) == need_histogram


@pytest.mark.parametrize("dtype", [np.uint8, np.uint16, np.float32])
@pytest.mark.parametrize("need_histogram", [False, True])
def test_minmax(dtype: np.dtype, need_histogram: bool) -> None:
    data = np.array([10, 20, 30, 40, 50], dtype=dtype).reshape(1, -1)
    stats = compute_image_stats(data, ClimsMinMax(), need_histogram=need_histogram)
    assert stats.clims[0] == pytest.approx(10, abs=1)
    assert stats.clims[1] == pytest.approx(50, abs=1)
    assert (stats.counts is not None) == need_histogram


@pytest.mark.parametrize(
    "data, policy, abs_tol, sig_bits",
    [
        (
            RNG.integers(0, 4095, (512, 512), dtype=np.uint16),
            ClimsPercentile(min_percentile=1, max_percentile=99),
            50,
            12,
        ),
        (
            RNG.random((100, 100), dtype=np.float32),
            ClimsPercentile(min_percentile=5, max_percentile=95),
            0.05,
            None,
        ),
        (
            np.array([-5, -1, 0, 3], dtype=np.int16),
            ClimsPercentile(min_percentile=1, max_percentile=99),
            1.0,
            None,
        ),
    ],
    ids=["uint16-12bit", "float32", "int16-negative"],
)
def test_percentile(
    data: np.ndarray,
    policy: ClimsPercentile,
    abs_tol: float,
    sig_bits: int | None,
) -> None:
    kwargs = {"significant_bits": sig_bits} if sig_bits else {}
    stats = compute_image_stats(data, policy, need_histogram=False, **kwargs)
    ref = tuple(np.nanpercentile(data, [policy.min_percentile, policy.max_percentile]))
    assert stats.clims[0] == pytest.approx(ref[0], abs=abs_tol)
    assert stats.clims[1] == pytest.approx(ref[1], abs=abs_tol)


@pytest.mark.parametrize(
    "data, abs_tol",
    [
        (RNG.integers(0, 255, (100, 100), dtype=np.uint8), 5),
        (RNG.random((100, 100), dtype=np.float32), 0.01),
    ],
    ids=["uint8", "float32"],
)
def test_stddev(data: np.ndarray, abs_tol: float) -> None:
    stats = compute_image_stats(data, ClimsStdDev(n_stdev=2), need_histogram=False)
    ref_mean = float(np.nanmean(data))
    ref_std = float(np.nanstd(data))
    assert stats.clims[0] == pytest.approx(ref_mean - 2 * ref_std, abs=abs_tol)
    assert stats.clims[1] == pytest.approx(ref_mean + 2 * ref_std, abs=abs_tol)


@pytest.mark.parametrize(
    "data, expected_clims",
    [
        (np.full((10, 10), 42, dtype=np.uint8), (42, 42)),
        (np.full((10, 10), 3.14, dtype=np.float32), (3.14, 3.14)),
    ],
    ids=["uint8", "float32"],
)
def test_single_value_data(
    data: np.ndarray, expected_clims: tuple[float, float]
) -> None:
    stats = compute_image_stats(data, ClimsMinMax(), need_histogram=True)
    assert stats.clims == pytest.approx(expected_clims, abs=1)
    assert stats.counts is not None


def test_all_nan_float() -> None:
    data = np.full((10, 10), np.nan, dtype=np.float32)
    with pytest.warns(RuntimeWarning):
        stats = compute_image_stats(data, ClimsMinMax(), need_histogram=False)
    assert np.isnan(stats.clims[0])


def test_significant_bits_passed_through() -> None:
    """12-bit significant_bits should produce 4096-bin histogram."""
    data = np.array([[0, 2000, 4000]], dtype=np.uint16)
    stats = compute_image_stats(
        data, ClimsMinMax(), need_histogram=True, significant_bits=12
    )
    assert stats.counts is not None
    assert stats.counts.shape == (4096,)


@pytest.mark.parametrize(
    "dtype",
    [
        np.float16,
        np.float32,
        np.float64,
        np.int8,
        np.int16,
        np.int32,
        np.int64,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint64,
    ],
)
def test_calc_hist_bins_float32(dtype) -> None:
    """_compute_histogram should return valid counts and edges for various dtypes."""
    if np.issubdtype(dtype, np.floating):
        data = np.random.rand(1000).astype(dtype) * 100
    else:
        data = np.random.randint(0, 100, size=(1000,), dtype=dtype)
    counts, edges = _compute_histogram(data, significant_bits=None)
    assert len(counts) > 0
    assert len(edges) == len(counts) + 1



from __future__ import annotations

from unittest.mock import MagicMock

import cmap
import wx
from pytest import fixture

from ndv.models._lut_model import ClimsManual, ClimsMinMax, ClimsPercentile, LUTModel
from ndv.views._wx._array_view import WxLUTView
from ndv.views.bases._graphics._canvas import HistogramCanvas


@fixture
def model() -> LUTModel:
    return LUTModel()


@fixture
def view(wxapp: wx.App, model: LUTModel) -> WxLUTView:
    # NB: wx.App necessary although unused
    frame = wx.Frame(None)
    view = WxLUTView(frame)
    assert view.model is None
    view.model = model
    assert view.model is model
    return view


def test_WxLUTView_update_model(model: LUTModel, view: WxLUTView) -> None:
    """Ensures the view updates when the model is changed."""

    # Test modifying model.clims
    assert view._wxwidget.auto_clim.GetValue()
    model.clims = ClimsManual(min=0, max=1)
    assert not view._wxwidget.auto_clim.GetValue()
    model.clims = ClimsPercentile(min_percentile=0, max_percentile=100)
    assert view._wxwidget.auto_clim.GetValue()
    model.clims = ClimsPercentile(min_percentile=1, max_percentile=99)
    assert view._wxwidget.lower_tail.GetValue() == 1
    assert view._wxwidget.upper_tail.GetValue() == 1

    # Test modifying model.visible
    assert view._wxwidget.visible.GetValue()
    model.visible = False
    assert not view._wxwidget.visible.GetValue()
    model.visible = True
    assert view._wxwidget.visible.GetValue()

    # Test modifying model.cmap
    new_cmap = cmap.Colormap("red")
    assert view._wxwidget.cmap.GetValue() != new_cmap
    model.cmap = new_cmap
    assert view._wxwidget.cmap.GetValue() == new_cmap


def test_WxLUTView_update_view(wxapp: wx.App, model: LUTModel, view: WxLUTView) -> None:
    """Ensures the model updates when the view is changed."""

    def processEvent(evt: wx.PyEventBinder | int, wdg: wx.Window) -> None:
        ev = wx.PyCommandEvent(evt.typeId, wdg.GetId())
        wx.PostEvent(wdg.GetEventHandler(), ev)
        # Borrowed from:
        # https://github.com/wxWidgets/Phoenix/blob/master/unittests/wtc.py#L41
        evtLoop = wxapp.GetTraits().CreateEventLoop()
        wx.EventLoopActivator(evtLoop)
        evtLoop.YieldFor(wx.EVT_CATEGORY_ALL)

    new_clims = (5, 6)
    assert model.clims != new_clims
    clim_wdg = view._wxwidget.clims
    clim_wdg.SetValue(*new_clims)
    processEvent(wx.EVT_SLIDER, clim_wdg)
    assert model.clims == ClimsManual(min=5, max=6)

    new_visible = not model.visible
    vis_wdg = view._wxwidget.visible
    vis_wdg.SetValue(new_visible)
    processEvent(wx.EVT_CHECKBOX, vis_wdg)
    assert model.visible == new_visible

    new_cmap = cmap.Colormap("red")
    assert model.cmap != new_cmap
    cmap_wdg = view._wxwidget.cmap
    cmap_wdg.SetValue(new_cmap.name)
    processEvent(wx.EVT_COMBOBOX, cmap_wdg)
    assert model.cmap == new_cmap

    # Test toggling auto_clim
    auto_wdg = view._wxwidget.auto_clim
    auto_wdg.SetValue(True)
    processEvent(wx.EVT_TOGGLEBUTTON, auto_wdg)
    assert model.clims == ClimsPercentile(min_percentile=0, max_percentile=100)
    auto_wdg.SetValue(False)
    processEvent(wx.EVT_TOGGLEBUTTON, auto_wdg)
    mi, ma = view._wxwidget.clims.GetValues()
    assert model.clims == ClimsManual(min=mi, max=ma)

    # Test modifying tails changes percentiles
    auto_wdg.SetValue(True)
    processEvent(wx.EVT_TOGGLEBUTTON, auto_wdg)
    assert model.clims == ClimsPercentile(min_percentile=0, max_percentile=100)
    lower_wdg = view._wxwidget.lower_tail
    lower_wdg.SetValue(0.1)
    processEvent(wx.EVT_SPINCTRLDOUBLE, lower_wdg)
    assert model.clims == ClimsPercentile(min_percentile=0.1, max_percentile=100)
    upper_wdg = view._wxwidget.upper_tail
    upper_wdg.SetValue(0.2)
    processEvent(wx.EVT_SPINCTRLDOUBLE, upper_wdg)
    assert model.clims == ClimsPercentile(min_percentile=0.1, max_percentile=99.8)

    # When gui clims change, autoscale should be disabled
    model.clims = ClimsMinMax()
    clim_wdg = view._wxwidget.clims
    clim_wdg.SetValue(0, 1)
    processEvent(wx.EVT_SLIDER, clim_wdg)
    assert model.clims == ClimsManual(min=0, max=1)


def test_WxLUTView_histogram_controls(wxapp: wx.App, view: WxLUTView) -> None:
    def processEvent(evt: wx.PyEventBinder | int, wdg: wx.Window) -> None:
        ev = wx.PyCommandEvent(evt.typeId, wdg.GetId())
        wx.PostEvent(wdg.GetEventHandler(), ev)
        # Borrowed from:
        # https://github.com/wxWidgets/Phoenix/blob/master/unittests/wtc.py#L41
        evtLoop = wxapp.GetTraits().CreateEventLoop()
        wx.EventLoopActivator(evtLoop)
        evtLoop.YieldFor(wx.EVT_CATEGORY_ALL)

    # Mock up a histogram
    hist_mock = MagicMock(spec=HistogramCanvas)
    # Note that containing the frontend widget within a frame prevents segfaults
    frame = wx.Frame(None)
    hist_frontend = wx.Window(frame)
    hist_mock.frontend_widget.return_value = hist_frontend

    # Add the histogram and assert it was correctly added
    view._add_histogram(hist_mock)
    assert view.histogram is hist_mock

    # Assert histogram button toggles visibility
    hist_btn = view._wxwidget.histogram_btn
    log_wdg = view._wxwidget.log_btn
    reset_wdg = view._wxwidget.set_hist_range_btn

    hist_btn.SetValue(True)
    processEvent(wx.EVT_TOGGLEBUTTON, hist_btn)
    assert log_wdg.IsShown()
    assert reset_wdg.IsShown()
    assert hist_frontend.IsShown()

    hist_btn.SetValue(False)
    processEvent(wx.EVT_TOGGLEBUTTON, hist_btn)
    assert not log_wdg.IsShown()
    assert not reset_wdg.IsShown()
    assert not hist_frontend.IsShown()

    # Assert toggling the log button alters the logarithmic base
    log_wdg.SetValue(True)
    processEvent(wx.EVT_TOGGLEBUTTON, log_wdg)
    hist_mock.set_log_base.assert_called_once_with(10)
    hist_mock.reset_mock()

    log_wdg.SetValue(False)
    processEvent(wx.EVT_TOGGLEBUTTON, log_wdg)
    hist_mock.set_log_base.assert_called_once_with(None)
    hist_mock.reset_mock()

    # Assert pressing the reset view button sets the histogram range
    processEvent(wx.EVT_BUTTON, reset_wdg)
    hist_mock.set_range.assert_called_once_with()
    hist_mock.reset_mock()

    # Assert pressing the reset view button turns off log mode
    log_wdg.SetValue(True)
    processEvent(wx.EVT_TOGGLEBUTTON, log_wdg)
    processEvent(wx.EVT_BUTTON, reset_wdg)
    assert not log_wdg.GetValue()
    hist_mock.reset_mock()



from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator, MutableMapping
from contextlib import suppress
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Protocol,
    TypeVar,
    cast,
    get_args,
    overload,
)

from psygnal import Signal
from pydantic import TypeAdapter
from pydantic_core import SchemaValidator, core_schema

if TYPE_CHECKING:
    from pydantic import GetCoreSchemaHandler

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")
_VT_co = TypeVar("_VT_co", covariant=True)
_NULL = object()


class SupportsKeysAndGetItem(Protocol[_KT, _VT_co]):
    def keys(self) -> Iterable[_KT]: ...
    def __getitem__(self, key: _KT, /) -> _VT_co: ...


class ValidatedEventedDict(MutableMapping[_KT, _VT]):
    item_added = Signal(str, object)  # key, new_value
    item_removed = Signal(str, object)  # key, old_value
    item_changed = Signal(str, object, object)  # key, new_value, old_value
    value_changed = Signal()

    # long ugly overloads to support all possible ways to initialize a ValidatedDict
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(  # type: ignore[misc]
        self: dict,
        key_validator: Callable[[Any], _KT] | None = None,
        value_validator: Callable[[Any], _VT] | None = None,
        **kwargs: _VT,
    ) -> None: ...
    @overload
    def __init__(
        self,
        map: SupportsKeysAndGetItem[_KT, _VT],
        /,
        key_validator: Callable[[Any], _KT] | None = None,
        value_validator: Callable[[Any], _VT] | None = None,
    ) -> None: ...
    @overload
    def __init__(  # type: ignore[misc]
        self: dict,
        map: SupportsKeysAndGetItem[str, _VT],
        /,
        key_validator: Callable[[Any], _KT] | None = None,
        value_validator: Callable[[Any], _VT] | None = ...,
        validate_lookup: bool = ...,
        **kwargs: _VT,
    ) -> None: ...
    @overload
    def __init__(
        self,
        iterable: Iterable[tuple[_KT, _VT]],
        /,
        key_validator: Callable[[Any], _KT] | None = ...,
        value_validator: Callable[[Any], _VT] | None = ...,
        validate_lookup: bool = ...,
    ) -> None: ...
    @overload
    def __init__(  # type: ignore[misc]
        self: dict,
        iterable: Iterable[tuple[str, _VT]],
        /,
        key_validator: Callable[[Any], _KT] | None = None,
        value_validator: Callable[[Any], _VT] | None = ...,
        validate_lookup: bool = ...,
        **kwargs: _VT,
    ) -> None: ...
    def __init__(  # type: ignore[misc] # does not accept all possible overloads
        self,
        *args: Any,
        key_validator: Callable[[Any], _KT] | None = None,
        value_validator: Callable[[Any], _VT] | None = None,
        validate_lookup: bool = False,
        **kwargs: Any,
    ) -> None:
        self._key_validator = key_validator
        self._value_validator = value_validator
        self._validate_lookup = validate_lookup
        _d = {}
        for k, v in dict(*args, **kwargs).items():
            if self._key_validator is not None:
                k = self._key_validator(k)
            if self._value_validator is not None:
                v = self._value_validator(v)
            _d[k] = v
        self._dict: dict[_KT, _VT] = _d

    def __missing__(self, key: _KT) -> _VT:
        raise KeyError(key)

    # ---------------- abstract interface ----------------

    def __getitem__(self, key: _KT) -> _VT:
        if self._validate_lookup:
            key = self._validate_key(key)
        try:
            return self._dict[key]
        except KeyError:
            return self.__missing__(key)

    # def __setitem__(self, key: _KT, value: _VT) -> None:
    # we allow Any here because validation may change the type of the value.
    def __setitem__(self, key: Any, value: Any) -> None:
        key = self._validate_key(key)
        value = self._validate_value(value)
        before = self._dict.get(key, _NULL)
        self._dict[key] = value
        # if the value is the same as before, try to exit early without emitting signals
        # but catch exceptions that may be raised during __eq__ (like numpy)
        if before is not _NULL:
            with suppress(Exception):
                if before == value:
                    return
            self.item_changed.emit(key, value, before)
        else:
            self.item_added.emit(key, value)
        self.value_changed.emit()

    def __delitem__(self, key: _KT) -> None:
        if self._validate_lookup:
            key = self._validate_key(key)
        # TODO: maybe add removing signal (before actual removal) if needed
        item = self._dict.pop(key)
        self.item_removed.emit(key, item)
        self.value_changed.emit()

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self) -> Iterator[_KT]:
        return iter(self._dict)

    # batch operations, with a single value_changed  -------------------------------

    @overload
    def assign(self, m: SupportsKeysAndGetItem[_KT, _VT], /, **kwargs: _VT) -> None: ...
    @overload
    def assign(self, m: Iterable[tuple[_KT, _VT]], /, **kwargs: _VT) -> None: ...
    @overload
    def assign(self, **kwargs: _VT) -> None: ...
    def assign(self, *args: Any, **kwargs: _VT) -> None:  # type: ignore[misc]
        """Override state with the key/value pairs from the mapping or iterable.

        Similar to update, but clears the dictionary first (without signals), replacing
        the contents with the key/value pairs from the mapping or iterable, and then
        emitting a single value_changed signal at the end.
        """
        with self.value_changed.blocked():
            self.clear()
            self.update(*args, **kwargs)
        self.value_changed.emit()

    @overload
    def update(self, m: SupportsKeysAndGetItem[_KT, _VT], /, **kwargs: _VT) -> None: ...
    @overload
    def update(self, m: Iterable[tuple[_KT, _VT]], /, **kwargs: _VT) -> None: ...
    @overload
    def update(self, **kwargs: _VT) -> None: ...
    def update(self, *args: Any, **kwargs: _VT) -> None:  # type: ignore[misc]
        """Update the dictionary with the key/value pairs from the mapping or iterable.

        only emit a single value_changed signal at the end.
        """
        with self.value_changed.blocked():
            super().update(*args, **kwargs)
        # TODO: only emit if anything was caught
        self.value_changed.emit()

    def clear(self) -> None:
        """Clear the dictionary.

        only emit a single value_changed signal at the end.
        """
        with self.value_changed.blocked():
            super().clear()
        # TODO: only emit if anything was caught
        self.value_changed.emit()

    # -----------------------------------------------------

    @cached_property
    def _validate_key(self) -> Callable[[Any], _KT]:
        """Return a function that validates keys."""
        if self._key_validator is not None:
            return self._key_validator
        # No key validator was provided during init.  Try to get the key type from the
        # class type hint and return a validator function for it.
        # __orig_class__ is not available during __init__
        # https://discuss.python.org/t/runtime-access-to-type-parameters/37517
        cls = getattr(self, "__orig_class__", None) or type(self)
        if args := get_args(cls):
            return TypeAdapter(args[0]).validator.validate_python
        # fall back to identity function
        return lambda x: x

    @cached_property
    def _validate_value(self) -> Callable[[Any], _VT]:
        """Return a function that validates values."""
        if self._value_validator is not None:
            return self._value_validator
        # No value validator was provided during init.  Try to get the value type from
        # the class type hint and return a validator function for it.
        # __orig_class__ is not available during __init__
        # https://discuss.python.org/t/runtime-access-to-type-parameters/37517
        cls = getattr(self, "__orig_class__", None) or type(self)
        if len(args := get_args(cls)) > 1:
            return TypeAdapter(args[1]).validator.validate_python
        # fall back to identity function
        return lambda x: x

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._dict!r})"

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        """Return the Pydantic core schema for this object.

        In this method, we parse the key and value types of the `source_type`, which
        be something like ValidatedDict[KT, VT]. And then create a validator function
        that creates a new instance of the ValidatedDict during assignment, passing in
        the key and value validator functions (from pydantic).

        Parameters
        ----------
        source_type : Any
            The source type.  This will usually be `cls`.
        handler : GetCoreSchemaHandler
            Handler to call into the next CoreSchema schema generation function.
        """
        # get key/value types from the source_type type hint.
        key_type = val_type = Any
        if args := get_args(source_type):
            key_type = args[0]
            if len(args) > 1:
                val_type = args[1]

        # get key/value core schemas for the key/value types.
        keys_schema = _get_schema(key_type, handler)
        values_schema = _get_schema(val_type, handler)
        validate_key = SchemaValidator(keys_schema).validate_python
        validate_value = SchemaValidator(values_schema).validate_python

        # define function that creates new instance during assignment
        # passing in the validator functions.
        def _new(*args: Any, **kwargs: Any) -> ValidatedEventedDict[_KT, _VT]:
            return cls(  # type: ignore[call-overload,no-any-return]
                *args,
                key_validator=validate_key,
                value_validator=validate_value,
                **kwargs,
            )

        # this schema for this validated dict
        dict_schema = core_schema.dict_schema(
            keys_schema=keys_schema,
            values_schema=values_schema,
        )
        # wrap the schema with a validator function that creates a new instance,
        # passing in the key/value validators.
        return core_schema.no_info_after_validator_function(
            function=_new,
            schema=dict_schema,
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda x: x._dict, return_schema=dict_schema
            ),
        )


def _get_schema(hint: Any, handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
    # check if the hint already has a core schema attached to it.
    if hasattr(hint, "__pydantic_core_schema__"):
        return cast("core_schema.CoreSchema", hint.__pydantic_core_schema__)
    # otherwise, call the handler to get the core schema.
    return handler.generate_schema(hint)



"""Custom hooks used in our documentation.

This is registered in the `hooks` section of `mkdocs.yml` file and is used to apply
all of our customizations to the documentation build process, such as automatically
generating screenshots from scripts.

Screenshots can be inserted into the documentation by using the following syntax:

```markdown
{{ screenshot: some/file/relative/to/repo_root.py }}
```
"""

from __future__ import annotations

import os
import re
import runpy
import subprocess
import sys
import time
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal
from unittest.mock import patch

import markdown
from mkdocs.plugins import get_plugin_logger
from mkdocs.structure.files import File, InclusionLevel
from qtpy.QtCore import QBuffer, QCoreApplication, QIODevice
from qtpy.QtWidgets import QApplication

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from matplotlib.dviread import Page
    from mkdocs.config import Config
    from mkdocs.config.defaults import MkDocsConfig
    from mkdocs.structure.files import Files


# logger for logging to mkdocs status
LOGGER = get_plugin_logger("ndv_docs")
# Path to the docs directory
DOCS = Path(__file__).parent
ROOT = DOCS.parent
# pattern to detect {{ screenshot: some/file.py }}
SCREENSHOT_RE = re.compile(r"{{\s*screenshot:\s*(.+?)\s*}}")
# a mapping of {hash -> File} for all screenshots we've generated
SCREENSHOTS: defaultdict[int, set[File]] = defaultdict(set)
GEN_SCREENSHOTS = os.getenv("GEN_SCREENSHOTS", "1") not in ("0", "false", "False")


def on_startup(command: Literal["build", "gh-deploy", "serve"], dirty: bool) -> None:
    """Runs once at the very beginning of an mkdocs invocation."""
    sys.path.append(str(DOCS))


def on_config(config: Config, **__: Any) -> None:
    """First event called on build, run immediately after user cfg is loaded/validated.

    Any alterations to the config should be made here.
    """
    config["markdown_extensions"].append("hooks")


def on_page_markdown(
    markdown: str, page: Page, config: MkDocsConfig, files: Files
) -> str:
    """Called after the page's markdown is loaded from file.

    Can be used to alter the Markdown source text. The meta- data has been stripped off
    and is available as page.meta at this point.
    """
    # ------------------------- Screenshot Generation ---------------------------------
    #

    def get_screenshot_link(match: re.Match) -> str:
        """Callback for SCREENSHOT_RE.sub.

        Return a markdown link to replace `{{ screenshot: some/file.py }}`
        """
        script_path = Path(match.group(1))
        # add the script to the watch list so that mkdocs will rebuild the page
        # if the script is edited
        abs_path = str(ROOT / script_path)
        if abs_path not in config.watch:
            config.watch.append(abs_path)

        # if we've never seen the content of this script before, generate a screenshot
        # and store it in the SCREENSHOTS dict
        if not (ss_files := SCREENSHOTS[hash(script_path.read_bytes())]):
            LOGGER.info(f"Generating screenshot for {script_path}")
            for content, mode in generate_screenshots(script_path):
                src_uri = f"screenshots/{script_path.stem}_{mode}.png"
                LOGGER.info(f"   >> {mode}: {src_uri}")

                # mkdocs takes care of copying the file to the site_dir
                # we just have to create a File object for it
                file = File.generated(
                    config,
                    src_uri=src_uri,
                    content=content,
                    inclusion=InclusionLevel.NOT_IN_NAV,
                )
                ss_files.add(file)

        link = ""
        for file in ss_files:
            # build markdown links, with lazy loading and theme awareness
            files.append(file)
            mode = "dark" if "dark" in file.src_uri else "light"
            alt_txt = f"Screenshot generated with {script_path.stem}"
            uri = f"../{file.src_uri}#only-{mode}"
            # note, the .auto-screenshot class is styled in ndv.css
            link += f"![{alt_txt}]({uri}){{ .auto-screenshot loading=lazy }}\n"
        return link

    # Find all {{ screenshot: some/file.py }},
    # generate a screenshot for each file and replace the tag with the image link
    # this generates two links: one for light mode and one for dark mode
    if GEN_SCREENSHOTS:
        markdown = SCREENSHOT_RE.sub(get_screenshot_link, markdown)

    # ---------------------------------------------------------------------------

    return markdown


# ---------------------------- ScreenShot Generation ----------------------------

Mode = Literal["light", "dark"]


def generate_screenshots(script: Path) -> Iterable[tuple[bytes, Mode]]:
    """Generate a screenshot from a script."""
    script = Path(script).resolve()
    if not script.is_file():
        raise FileNotFoundError(f"File not found: {script}")
    if script.suffix != ".py":
        raise ValueError(f"Invalid file type: {script}")

    # patch any calls to QApplication([]) in scripts, so they don't cause an
    # error if the Application is already created.
    original_new = QApplication.__new__

    def _start_app(*_: Any) -> QCoreApplication:
        if (app := QApplication.instance()) is None:
            QApplication._APP = app = original_new(*_)  # type: ignore[attr-defined]
            QApplication.__init__(app, [])  # type: ignore[arg-type]
        return app

    patch_app = patch.object(QApplication, "__new__", _start_app)

    # We'll collect yields from grab_top_widget into a list,
    # then yield them after the script has finished.
    results: list[tuple[bytes, Mode]] = []

    # patch QApplication.exec to grab screenshots instead of running the event loop

    def patched_exec(_: Any) -> int:
        # Call the generator and store items in results
        for item in grab_top_widget():
            results.append(item)
        # Return 0, or whatever exit code you like, so that
        # QApp thinks exec() ended
        return 0

    patch_exec = patch.object(QApplication, "exec", patched_exec)
    with patch_app, patch_exec:
        # run the script
        runpy.run_path(str(script), run_name="__main__")

    # return the screenshot grabbed by the grab_top_widget function
    return results


def grab_top_widget(
    fmt: str = "png",
) -> Iterator[tuple[bytes, Literal["light", "dark"]]]:
    """Find the top widget and return bytes containing a screenshot."""
    from qtpy.QtWidgets import QFrame

    try:
        main_wdg = next(
            wdg for wdg in QApplication.topLevelWidgets() if not isinstance(wdg, QFrame)
        )
    except StopIteration:
        return

    QApplication.sendPostedEvents()
    # the sleep here is mostly to allow the widget to render,
    # the biggest issue at the time is the animated expansion of the luts
    # QcollapsibleWidget.  That should be fixed at superqt to not animate if desired.
    for _i in range(10):
        time.sleep(0.1)
        QApplication.processEvents()

    def _get_bytes() -> bytes:
        buffer = QBuffer()
        buffer.open(QIODevice.OpenModeFlag.WriteOnly)
        pixmap = main_wdg.grab()
        pixmap.save(buffer, fmt)
        return buffer.data().data()

    modes: list[Mode] = ["dark", "light"] if sys.platform == "darwin" else ["light"]
    for mode in modes:
        set_dark_mode(mode == "dark")
        for _i in range(12):
            time.sleep(0.1)
            QApplication.processEvents()
        yield _get_bytes(), mode

    for wdg in QApplication.topLevelWidgets():
        wdg.close()
        wdg.deleteLater()
    QApplication.processEvents()
    QApplication.processEvents()


def set_dark_mode(bool: bool) -> bool:
    if not sys.platform == "darwin":
        return False
    try:
        subprocess.check_call(
            [
                "osascript",
                "-e",
                'tell application "System Events" to tell appearance preferences '
                f"to set dark mode to {bool}",
            ]
        )
        return True
    except subprocess.CalledProcessError:
        return False


# ---------------------------- Custom Markdown Extension ----------------------------

# this stuff can be removed if we stop using snippets to insert markdown
# (like README.md) that has github style admonitions.


class GhAdmonitionPreprocessor(markdown.preprocessors.Preprocessor):
    """Convert github-style admonitions to MkDocs-style admonitions."""

    def convert_gh_admonitions(self, lines: list[str]) -> list[str]:
        """Transforms [!X] blocks in Markdown into !!! X blocks."""
        output_lines = []
        inside_note_block = False
        type_ = None
        note_content = []

        for line in lines:
            if line.startswith("> [!"):  # Start of a note block
                inside_note_block = True
                type_ = line.split("[!", 1)[1].split("]", 1)[0].lower()
            elif inside_note_block and line.startswith(
                "> "
            ):  # Content of the note block
                note_content.append(line[2:])  # Strip the '> ' prefix
            else:
                if inside_note_block:  # End of the note block
                    output_lines.append(f"!!! {type_}")
                    output_lines.extend(
                        f"    {content_line}" for content_line in note_content
                    )
                    inside_note_block = False
                    type_ = None
                    note_content = []

                output_lines.append(line)  # Add non-note lines as is

        # If the file ends with a note block, finalize it
        if inside_note_block:
            output_lines.append(f"!!! {type_}")
            output_lines.extend(f"    {content_line}" for content_line in note_content)

        return output_lines

    def run(self, lines: list[str]) -> list[str]:
        return self.convert_gh_admonitions(lines)


def makeExtension() -> markdown.Extension:
    class MyExtension(markdown.Extension):
        def extendMarkdown(self, md: markdown.Markdown) -> None:
            # must be < 32 which is what pymdownx.snippets uses
            md.preprocessors.register(GhAdmonitionPreprocessor(), "snippet-post", 2)

    return MyExtension()



from __future__ import annotations

from pydantic import field_validator

from ndv.models._base_model import NDVModel


class RectangularROIModel(NDVModel):
    """Representation of an axis-aligned rectangular Region of Interest (ROI).

    Attributes
    ----------
    visible : bool
        Whether to display this roi.
    bounding_box : tuple[tuple[float, float], tuple[float, float]]
        The minimum (2D) point and the maximum (2D) point contained within the
        region. Using these two points, an axis-aligned bounding box can be
        constructed.
    """

    visible: bool = True
    bounding_box: tuple[tuple[float, float], tuple[float, float]] = ((0, 0), (0, 0))

    @field_validator("bounding_box", mode="after")
    @classmethod
    def _validate_bounding_box(
        cls, bb: tuple[tuple[float, float], tuple[float, float]]
    ) -> tuple[tuple[float, float], tuple[float, float]]:
        x1 = min(bb[0][0], bb[1][0])
        y1 = min(bb[0][1], bb[1][1])
        x2 = max(bb[0][0], bb[1][0])
        y2 = max(bb[0][1], bb[1][1])
        return ((x1, y1), (x2, y2))



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from ._view_base import Viewable

if TYPE_CHECKING:
    import cmap

    from ndv.models._lut_model import ClimPolicy, LUTModel


class LUTView(Viewable):
    """Manages LUT properties (contrast, colormap, etc...) in a view object."""

    _model: LUTModel | None = None
    _fallback_name: str = ""

    @abstractmethod
    def set_channel_name(self, name: str) -> None:
        """Set the name of the channel to `name`."""

    def set_fallback_name(self, name: str) -> None:
        """Set the data-derived or default fallback name for this channel.

        The display name resolves as: `lut.name` if set, else `fallback_name`.
        """
        self._fallback_name = name
        self._refresh_display_name()

    def _refresh_display_name(self) -> None:
        """Resolve and apply the display name from model.name and fallback."""
        if self._model and self._model.name:
            name = self._model.name
        else:
            name = self._fallback_name
        self.set_channel_name(name)

    @abstractmethod
    def set_clim_policy(self, policy: ClimPolicy) -> None:
        """Set the clim policy to `policy`.

        Usually corresponds to an "autoscale" checkbox.

        Note that this method must not modify the backing LUTModel.
        """

    @abstractmethod
    def set_colormap(self, cmap: cmap.Colormap) -> None:
        """Set the colormap to `cmap`.

        Usually corresponds to a dropdown menu.

        Note that this method must not modify the backing LUTModel.
        """

    @abstractmethod
    def set_clims(self, clims: tuple[float, float]) -> None:
        """Set the (low, high) contrast limits to `clims`.

        Usually this will be a range slider or two text boxes.

        Note that this method must not modify the backing LUTModel.
        """

    def set_clim_bounds(
        self,
        bounds: tuple[float | None, float | None] = (None, None),
    ) -> None:
        """Defines the minimum and maximum possible values for the clims."""

    @abstractmethod
    def set_channel_visible(self, visible: bool) -> None:
        """Check or uncheck the visibility indicator of the LUT.

        Usually corresponds to a checkbox.

        Note that this method must not modify the backing LUTModel.
        """

    def set_gamma(self, gamma: float) -> None:
        """Set the gamma value of the LUT.

        Note that this method must not modify the backing LUTModel.
        """
        return None

    @property
    def model(self) -> LUTModel | None:
        return self._model

    @model.setter
    def model(self, model: LUTModel | None) -> None:
        # Disconnect old model
        if self._model is not None:
            self._model.events.clims.disconnect(self.set_clim_policy)
            self._model.events.clim_bounds.disconnect(self.set_clim_bounds)
            self._model.events.cmap.disconnect(self.set_colormap)
            self._model.events.gamma.disconnect(self.set_gamma)
            self._model.events.visible.disconnect(self.set_channel_visible)
            self._model.events.name.disconnect(self._refresh_display_name)

        # Connect new model
        self._model = model
        if self._model is not None:
            self._model.events.clims.connect(self.set_clim_policy)
            self._model.events.clim_bounds.connect(self.set_clim_bounds)
            self._model.events.cmap.connect(self.set_colormap)
            self._model.events.gamma.connect(self.set_gamma)
            self._model.events.visible.connect(self.set_channel_visible)
            self._model.events.name.connect(self._refresh_display_name)

        self.synchronize()

    def synchronize(self) -> None:
        """Aligns the view against the backing model."""
        if model := self._model:
            self.set_clim_policy(model.clims)
            self.set_clim_bounds(model.clim_bounds)
            self.set_colormap(model.cmap)
            self.set_gamma(model.gamma)
            self.set_channel_visible(model.visible)
        self._refresh_display_name()



from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Annotated, Any, Literal, TypeAlias

import numpy as np
import numpy.typing as npt
from annotated_types import Gt, Interval
from cmap import Colormap
from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    model_validator,
)

from ._base_model import NDVModel

AutoscaleType: TypeAlias = (
    Callable[[npt.ArrayLike], tuple[float, float]] | tuple[float, float] | bool
)


class ClimPolicy(BaseModel, ABC):
    """ABC for contrast limit policies."""

    model_config = ConfigDict(frozen=True, extra="forbid")

    @abstractmethod
    def get_limits(self, image: npt.NDArray) -> tuple[float, float]:
        """Return the contrast limits for the given image."""

    @property
    def is_manual(self) -> bool:
        return self.__class__ == ClimsManual


class ClimsManual(ClimPolicy):
    """Manually specified contrast limits.

    Attributes
    ----------
    min: float
        The minimum contrast limit.
    max: float
        The maximum contrast limit.
    """

    clim_type: Literal["manual"] = "manual"
    min: float
    max: float

    def get_limits(self, data: npt.NDArray) -> tuple[float, float]:
        return self.min, self.max

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, ClimsManual)
            and abs(self.min - other.min) < 1e-6
            and abs(self.max - other.max) < 1e-6
        )


class ClimsMinMax(ClimPolicy):
    """Autoscale contrast limits based on the minimum and maximum values in the data."""

    clim_type: Literal["minmax"] = "minmax"

    def get_limits(self, data: npt.NDArray) -> tuple[float, float]:
        return (np.nanmin(data), np.nanmax(data))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ClimsMinMax)


class ClimsPercentile(ClimPolicy):
    """Autoscale contrast limits based on percentiles of the data.

    Attributes
    ----------
    min_percentile: float
        The lower percentile for the contrast limits.
    max_percentile: float
        The upper percentile for the contrast limits.
    """

    clim_type: Literal["percentile"] = "percentile"
    min_percentile: Annotated[float, Interval(ge=0, le=100)] = 0
    max_percentile: Annotated[float, Interval(ge=0, le=100)] = 100

    def get_limits(self, data: npt.NDArray) -> tuple[float, float]:
        return tuple(np.nanpercentile(data, [self.min_percentile, self.max_percentile]))

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, ClimsPercentile)
            and self.min_percentile == other.min_percentile
            and self.max_percentile == other.max_percentile
        )


class ClimsStdDev(ClimPolicy):
    """Automatically set contrast limits based on standard deviations from the mean.

    Attributes
    ----------
    n_stdev: float
        Number of standard deviations to use.
    center: Optional[float]
        Center value for the standard deviation calculation. If None, the mean is
        used.
    """

    clim_type: Literal["stddev"] = "stddev"
    n_stdev: Annotated[float, Gt(0)] = 2  # number of standard deviations
    center: float | None = None  # None means center around the mean

    def get_limits(self, data: npt.NDArray) -> tuple[float, float]:
        center = float(np.nanmean(data) if self.center is None else self.center)
        diff = float(self.n_stdev * np.nanstd(data))
        return center - diff, center + diff

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, ClimsStdDev)
            and self.n_stdev == other.n_stdev
            and self.center == other.center
        )


# we can add this, but it needs to have a proper pydantic serialization method
# similar to ReducerType
# class CustomClims(ClimPolicy):
#     type_: Literal["custom"] = "custom"
#     func: Callable[[npt.ArrayLike], tuple[float, float]]

#     def get_limits(self, data: npt.NDArray) -> tuple[float, float]:
#         return self.func(data)


def _validate_clims(v: Any) -> Any:
    if isinstance(v, str):
        v = {"clim_type": v}
    if v is None or (
        isinstance(v, dict)
        and v.get("min_percentile") == 0
        and v.get("max_percentile") == 100
    ):
        return ClimsMinMax()
    if isinstance(v, (tuple, list, np.ndarray)):
        if len(v) == 2:
            return ClimsManual(min=v[0], max=v[1])
        raise ValueError("Clims sequence must have exactly 2 elements.")
    return v


ClimsType: TypeAlias = Annotated[
    ClimsManual | ClimsPercentile | ClimsStdDev | ClimsMinMax,
    BeforeValidator(_validate_clims),
    # Field(discriminator="clim_type"),
]


class LUTModel(NDVModel):
    """Representation of how to display a channel of an array.

    Attributes
    ----------
    name : str
        Display name for this channel. Empty string means no explicit name.
    visible : bool
        Whether to display this channel.
        NOTE: This has implications for data retrieval, as we may not want to request
        channels that are not visible.
        See [`ArrayDisplayModel.current_index`][ndv.models.ArrayDisplayModel].
    cmap : cmap.Colormap
        [`cmap.Colormap`](https://cmap-docs.readthedocs.io/colormaps/) to use for this
        channel.  This can be expressed as any channel.  This can be expressed as any
        ["colormap like" object](https://cmap-docs.readthedocs.io/en/latest/colormaps/#colormaplike-objects)
    clims : Union[ClimsManual, ClimsPercentile, ClimsStdDev, ClimsMinMax]
        Method for determining the contrast limits for this channel.  If a 2-element
        `tuple` or `list` is provided, it is interpreted as a manual contrast limit.
    bounds : tuple[float | None, float | None]
        Optional extrema for limiting the range of the contrast limits
    gamma : float
        Gamma applied to the data before applying the colormap. By default, `1.0`.
    """

    name: str = ""
    visible: bool = True
    cmap: Colormap = Field(default_factory=lambda: Colormap("gray"))
    clims: ClimsType = Field(default_factory=ClimsMinMax)
    clim_bounds: tuple[float | None, float | None] = (None, None)
    gamma: float = 1.0

    @model_validator(mode="before")
    def _validate_model(cls, v: Any) -> Any:
        # cast bare string/colormap inputs to cmap declaration
        if isinstance(v, (str, Colormap)):
            return {"cmap": v}
        return v



from enum import Enum
from typing import TYPE_CHECKING

import cmap
import cmap._colormap
from pydantic import Field

from ndv.models._base_model import NDVModel

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import TypedDict

    from psygnal import Signal, SignalGroup

    class ArrayViewerModelKwargs(TypedDict, total=False):
        """Keyword arguments for `ArrayViewerModel`."""

        default_luts: "Sequence[cmap._colormap.ColorStopsLike]"
        interaction_mode: "InteractionMode"
        show_controls: bool
        show_3d_button: bool
        show_histogram_button: bool
        show_reset_zoom_button: bool
        show_roi_button: bool
        show_channel_mode_selector: bool
        show_play_button: bool
        show_data_info: bool
        show_progress_spinner: bool


class InteractionMode(str, Enum):
    """An enum defining graphical interaction mechanisms with an array Viewer."""

    PAN_ZOOM = "pan_zoom"
    CREATE_ROI = "create_roi"

    def __str__(self) -> str:
        """Return the string representation of the enum value."""
        return self.value


def _default_luts() -> "list[cmap.Colormap]":
    return [
        cmap.Colormap(x)
        for x in ("gray", "green", "magenta", "cyan", "red", "blue", "yellow")
    ]


class ArrayViewerModel(NDVModel):
    """Options and state for the [`ArrayViewer`][ndv.ArrayViewer].

    Attributes
    ----------
    interaction_mode : InteractionMode
        Describes the current interaction mode of the Viewer.
    show_controls : bool, optional
        Control visibility of *all* controls at once. By default True.
    show_3d_button : bool, optional
        Whether to show the 3D button, by default True.
    show_histogram_button : bool, optional
        Whether to show the histogram button, by default True.
    show_reset_zoom_button : bool, optional
        Whether to show the reset zoom button, by default True.
    show_roi_button : bool, optional
        Whether to show the ROI button, by default True.
    show_channel_mode_selector : bool, optional
        Whether to show the channel mode selector, by default True.
    show_play_button : bool, optional
        Whether to show the play button, by default True.
    show_progress_spinner : bool, optional
        Whether to show the progress spinner, by default
    show_data_info : bool, optional
        Whether to show shape, dtype, size, etc. about the array
    default_luts : list[cmap.Colormap], optional
        List of colormaps to use when populating the LUT dropdown menu in the viewer.
        Only editable upon initialization. Values may be any `cmap`
        [ColormapLike](https://cmap-docs.readthedocs.io/en/stable/colormaps/#colormaplike-objects)
        object (most commonly, just a string name of the colormap, like
        "gray" or "viridis").
    """

    interaction_mode: InteractionMode = InteractionMode.PAN_ZOOM
    show_controls: bool = True
    show_3d_button: bool = True
    show_histogram_button: bool = True
    show_reset_zoom_button: bool = True
    show_roi_button: bool = True
    show_channel_mode_selector: bool = True
    show_play_button: bool = True
    show_data_info: bool = True
    show_progress_spinner: bool = False
    default_luts: list[cmap.Colormap] = Field(
        default_factory=_default_luts, frozen=True
    )

    if TYPE_CHECKING:
        # just to make IDE autocomplete better
        # it's still hard to indicate dynamic members in the events group
        class ArrayViewerModelEvents(SignalGroup):
            """Signal group for ArrayViewerModel."""

            interaction_mode = Signal(InteractionMode, InteractionMode)
            show_controls = Signal(bool, bool)
            show_3d_button = Signal(bool, bool)
            show_histogram_button = Signal(bool, bool)
            show_reset_zoom_button = Signal(bool, bool)
            show_roi_button = Signal(bool, bool)
            show_channel_mode_selector = Signal(bool, bool)
            show_play_button = Signal(bool, bool)
            show_data_info = Signal(bool, bool)
            show_progress_spinner = Signal(bool, bool)

        events: ArrayViewerModelEvents = ArrayViewerModelEvents()  # type: ignore[misc]



from __future__ import annotations

from unittest.mock import MagicMock, Mock

import wx
from pytest import fixture

from ndv.models._viewer_model import ArrayViewerModel
from ndv.views._app import get_histogram_canvas_class
from ndv.views._wx._array_view import WxArrayView


@fixture
def viewer(wxapp: wx.App) -> WxArrayView:
    viewer = WxArrayView(MagicMock(), ArrayViewerModel())
    viewer.add_lut_view(None)
    return viewer


def _processEvent(
    wxapp: wx.App, evt: wx.PyEventBinder, wdg: wx.Control, **kwargs
) -> None:
    if evt == wx.EVT_ACTIVATE:
        active = kwargs.get("active", True)
        ev = wx.ActivateEvent(eventType=evt.typeId, active=active)
    else:
        ev = wx.PyCommandEvent(evt.typeId, wdg.GetId())

    wx.PostEvent(wdg.GetEventHandler(), ev)
    # Borrowed from:
    # https://github.com/wxWidgets/Phoenix/blob/master/unittests/wtc.py#L41
    evtLoop = wxapp.GetTraits().CreateEventLoop()
    wx.EventLoopActivator(evtLoop)
    evtLoop.YieldFor(wx.EVT_CATEGORY_ALL)


def test_array_options(viewer: WxArrayView) -> None:
    wxwdg = viewer._wxwidget
    wxwdg.Show()
    wxlut = viewer._luts[None]._wxwidget

    assert wxwdg.ndims_btn.IsShown()
    viewer._viewer_model.show_3d_button = False
    assert not wxwdg.ndims_btn.IsShown()

    assert wxlut.histogram_btn.IsShown()
    viewer._viewer_model.show_histogram_button = False
    assert not wxlut.histogram_btn.IsShown()

    assert wxwdg.set_range_btn.IsShown()
    viewer._viewer_model.show_reset_zoom_button = False
    assert not wxwdg.set_range_btn.IsShown()

    assert wxwdg.channel_mode_combo.IsShown()
    viewer._viewer_model.show_channel_mode_selector = False
    assert not wxwdg.channel_mode_combo.IsShown()

    assert wxwdg.add_roi_btn.IsShown()
    viewer._viewer_model.show_roi_button = False
    assert not wxwdg.add_roi_btn.IsShown()


def test_histogram(wxapp: wx.App, viewer: WxArrayView) -> None:
    channel = None
    lut = viewer._luts[channel]
    btn = lut._wxwidget.histogram_btn

    # Ensure lut signal gets passed through the viewer with the channel as the arg
    histogram_mock = Mock()
    viewer.histogramRequested.connect(histogram_mock)
    btn.SetValue(True)
    _processEvent(wxapp, wx.EVT_TOGGLEBUTTON, btn)
    histogram_mock.assert_called_once_with(channel)

    # Test adding the histogram widget puts it on the relevant lut
    assert len(lut._wxwidget._histogram_sizer.GetChildren()) == 1
    histogram = get_histogram_canvas_class()()  # will raise if not supported
    viewer.add_histogram(channel, histogram)
    assert len(lut._wxwidget._histogram_sizer.GetChildren()) == 2


# == Tests for display of channels ==


def test_display_options_visibility(wxapp: wx.App, viewer: WxArrayView) -> None:
    # display options button should appear only after thresh is reached
    # -2 to account for add_lut_view(None) in fixture
    ch = 0
    for ch in range(viewer._wxwidget._toolbar_display_thresh - 2):
        viewer.add_lut_view(ch)

    assert viewer._wxwidget.lut_selector.IsEnabled()

    assert not viewer._wxwidget.lut_selector.IsShown()
    assert not viewer._wxwidget._lut_toolbar_panel.IsShown()

    viewer.add_lut_view(ch + 1)

    assert viewer._wxwidget.lut_selector.IsShown()
    assert viewer._wxwidget._lut_toolbar_panel.IsShown()


def test_display_options_selection(wxapp: wx.App, viewer: WxArrayView) -> None:
    # display options button should appear after thresh reached
    num_channels = viewer._wxwidget._toolbar_display_thresh - 1
    for ch in range(num_channels):
        viewer.add_lut_view(ch)

    assert len(viewer._wxwidget.luts.Children) == len(viewer._luts)

    # all channels should initially be displayed
    for ch, lut_view in viewer._luts.items():
        if type(ch) is int or (type(ch) is str and ch.isdigit()):
            assert lut_view._wxwidget.IsShown()

    checklist = viewer._wxwidget.lut_selector._checklist

    # display off for a single channel
    checklist.Check(0, False)
    _processEvent(wxapp, wx.EVT_CHECKLISTBOX, checklist)

    assert not checklist.IsChecked(0)
    assert not viewer._luts[0]._wxwidget.IsShown()

    # display on again for a single channel
    checklist.Check(0, True)
    _processEvent(wxapp, wx.EVT_CHECKLISTBOX, checklist)

    assert checklist.IsChecked(0)
    assert viewer._luts[0]._wxwidget.IsShown()

    # Simulate what the controller does on mode change to grayscale:
    # it calls set_visible(False) on all multi-channel LUTs
    for ch, lut_view in viewer._luts.items():
        if type(ch) is int:
            lut_view.set_visible(False)
            assert not lut_view._wxwidget.IsShown()

    # Simulate switching back to composite: controller calls set_visible(True)
    # All previously-displayed channels should reappear
    for ch, lut_view in viewer._luts.items():
        if type(ch) is int:
            lut_view.set_visible(True)
            assert lut_view._wxwidget.IsShown(), (
                f"Channel {ch} should be visible after set_visible(True)"
            )

    # Now test interaction with display selector:
    # If a channel was deselected via the channel selector (set_display(False)),
    # it should NOT reappear after a mode round-trip
    viewer._luts[0].set_display(False)
    assert not viewer._luts[0]._wxwidget.IsShown()

    # Hide all (mode change to grayscale)
    for ch, lut_view in viewer._luts.items():
        if type(ch) is int:
            lut_view.set_visible(False)

    # Show all (mode change back to composite)
    for ch, lut_view in viewer._luts.items():
        if type(ch) is int:
            lut_view.set_visible(True)

    # Channel 0 was display-hidden, so it should stay hidden
    assert not viewer._luts[0]._wxwidget.IsShown()
    # Other channels should be visible
    assert viewer._luts[1]._wxwidget.IsShown()


def test_removed_channels(wxapp: wx.App, viewer: WxArrayView) -> None:
    # display options button should appear only after thresh is reached
    for ch in range(viewer._wxwidget._toolbar_display_thresh - 1):
        viewer.add_lut_view(ch)

    for ch in range(viewer._wxwidget._toolbar_display_thresh - 1):
        lut_view = viewer._luts[ch]
        viewer.remove_lut_view(lut_view)

    assert not viewer._wxwidget.lut_selector.IsShown()
    assert not viewer._wxwidget._lut_toolbar_panel.IsShown()

    # len == 1 to account for the None key
    assert len(viewer._luts) == 1
    assert len(viewer._wxwidget.luts.Children) == 1
    assert len(viewer._wxwidget.lut_selector._checklist.Children) == 0


def test_dropdown_popup(wxapp: wx.App, viewer: WxArrayView) -> None:
    for ch in range(viewer._wxwidget._toolbar_display_thresh - 1):
        viewer.add_lut_view(ch)

    assert not viewer._wxwidget.lut_selector._popup.IsShown()

    ch_selection_dropdown = viewer._wxwidget.lut_selector._dropdown_btn
    _processEvent(wxapp, wx.EVT_BUTTON, ch_selection_dropdown)

    assert viewer._wxwidget.lut_selector._popup.IsShown()

    _processEvent(
        wxapp, wx.EVT_ACTIVATE, viewer._wxwidget.lut_selector._popup, active=False
    )

    assert not viewer._wxwidget.lut_selector._popup.IsShown()


def test_none_all(wxapp: wx.App, viewer: WxArrayView) -> None:
    for ch in range(viewer._wxwidget._toolbar_display_thresh - 1):
        viewer.add_lut_view(ch)

    none_btn = viewer._wxwidget.lut_selector._select_none_btn
    all_btn = viewer._wxwidget.lut_selector._select_all_btn

    # select none
    _processEvent(wxapp, wx.EVT_BUTTON, none_btn)

    # all channels should be hidden
    for ch, lut_view in viewer._luts.items():
        if type(ch) is int or (type(ch) is str and ch.isdigit()):
            assert not lut_view._wxwidget.IsShown()

    # select all
    _processEvent(wxapp, wx.EVT_BUTTON, all_btn)

    # all channels should be displayed
    for ch, lut_view in viewer._luts.items():
        if type(ch) is int or (type(ch) is str and ch.isdigit()):
            assert lut_view._wxwidget.IsShown()


def test_key_event_filter(wxapp: wx.App) -> None:
    from ndv._types import KeyCode, KeyMod, KeyPressEvent
    from ndv.views._wx._app import WxAppWrap

    app = WxAppWrap()
    view = WxArrayView(MagicMock(), ArrayViewerModel())
    widget = view.frontend_widget()

    received: list[KeyPressEvent] = []
    view.keyPressed.connect(received.append)

    disconnect = app.filter_key_events(widget, view)

    # Simulate a Right arrow key press
    event = wx.KeyEvent(wx.wxEVT_CHAR_HOOK)
    event.SetKeyCode(wx.WXK_RIGHT)
    wx.PostEvent(widget.GetEventHandler(), event)
    evtLoop = wxapp.GetTraits().CreateEventLoop()
    wx.EventLoopActivator(evtLoop)
    evtLoop.YieldFor(wx.EVT_CATEGORY_ALL)

    assert len(received) == 1
    assert received[0].key == KeyCode.RIGHT
    assert received[0].mods == KeyMod.NONE

    disconnect()



# /// script
# dependencies = [
#     "ndv[pyqt,vispy]",
#     "sparse",
# ]
# ///
from __future__ import annotations

try:
    import sparse
except ImportError:
    raise ImportError("Please install sparse to run this example")

import numpy as np

import ndv

shape = (255, 4, 512, 512)
N = int(np.prod(shape) * 0.001)
coords = np.random.randint(low=0, high=shape, size=(N, len(shape)), dtype=np.uint16).T
data = np.random.randint(0, 255, N, dtype=np.uint8)


# Create the sparse array from the coordinates and data
sparse_array = sparse.COO(coords, data, shape=shape)

ndv.imshow(sparse_array)



from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING, Any, Literal, cast
from weakref import ReferenceType, WeakKeyDictionary, ref

import cmap as _cmap
import numpy as np
import pygfx
import pylinalg as la

from ndv._types import (
    CursorType,
    MouseButton,
    MouseMoveEvent,
    MousePressEvent,
    MouseReleaseEvent,
)
from ndv.models._viewer_model import ArrayViewerModel, InteractionMode
from ndv.views._app import filter_mouse_events
from ndv.views.bases import ArrayCanvas, CanvasElement, ImageHandle
from ndv.views.bases._graphics._canvas_elements import RectangularROIHandle, ROIMoveMode

from ._util import rendercanvas_class

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from pygfx.materials import ImageBasicMaterial
    from pygfx.resources import Texture


def _is_inside(bounding_box: np.ndarray | None, pos: Sequence[float]) -> bool:
    if bounding_box is None:
        return False
    return bool(
        bounding_box[0, 0] + 0.5 <= pos[0]
        and pos[0] <= bounding_box[1, 0] + 0.5
        and bounding_box[0, 1] + 0.5 <= pos[1]
        and pos[1] <= bounding_box[1, 1] + 0.5
    )


class PyGFXImageHandle(ImageHandle):
    def __init__(self, image: pygfx.Image | pygfx.Volume, render: Callable) -> None:
        self._image = image
        self._render = render
        self._grid = cast("Texture", image.geometry.grid)
        self._material = cast("ImageBasicMaterial", image.material)

    def data(self) -> np.ndarray:
        return self._grid.data  # type: ignore [no-any-return]

    def set_data(self, data: np.ndarray) -> None:
        # If dimensions are unchanged, reuse the buffer
        if data.shape == self._grid.data.shape:
            self._grid.data[:] = data  # pyright: ignore[reportOptionalSubscript]
            self._grid.update_range((0, 0, 0), self._grid.size)
        # Otherwise, the size (and maybe number of dimensions) changed
        # - we need a new buffer
        else:
            self._grid = pygfx.Texture(data, dim=2)
            self._image.geometry = pygfx.Geometry(grid=self._grid)
            # RGB images (i.e. 3D datasets) cannot have a colormap
            self._material.map = None if self._is_rgb() else self._cmap.to_pygfx()

    def visible(self) -> bool:
        return bool(self._image.visible)

    def set_visible(self, visible: bool) -> None:
        self._image.visible = visible
        self._render()

    def can_select(self) -> bool:
        return False

    def selected(self) -> bool:
        return False

    def set_selected(self, selected: bool) -> None:
        raise NotImplementedError("Images cannot be selected")

    def clims(self) -> Any:
        return self._material.clim

    def set_clims(self, clims: tuple[float, float]) -> None:
        self._material.clim = clims
        self._render()

    def gamma(self) -> float:
        return float(self._material.gamma)

    def set_gamma(self, gamma: float) -> None:
        self._material.gamma = gamma
        self._render()

    def colormap(self) -> _cmap.Colormap:
        return self._cmap

    def set_colormap(self, cmap: _cmap.Colormap) -> None:
        self._cmap = cmap
        # RGB (i.e. 3D) images should not have a colormap
        if not self._is_rgb():
            self._material.map = cmap.to_pygfx()
        self._render()

    def start_move(self, pos: Sequence[float]) -> None:
        pass

    def move(self, pos: Sequence[float]) -> None:
        pass

    def remove(self) -> None:
        if (par := self._image.parent) is not None:
            par.remove(self._image)

    def get_cursor(self, mme: MouseMoveEvent) -> CursorType | None:
        return None

    def _is_rgb(self) -> bool:
        return self.data().ndim == 3 and isinstance(self._image, pygfx.Image)


class PyGFXRectangle(RectangularROIHandle):
    def __init__(
        self,
        render: Callable,
        canvas_to_world: Callable,
        parent: pygfx.WorldObject | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        # Positional array backing visual objects
        # NB we need five points for the outline
        # The first and last rows should be identical
        self._positions: np.ndarray = np.zeros((5, 3), dtype=np.float32)

        # Visual objects
        self._fill = self._create_fill()
        self._outline = self._create_outline()

        # Handles used for ROI manipulation
        self._handle_rad = 5  # PIXELS
        self._handles = self._create_handles()

        # containing all ROI objects makes selection easier.
        self._container = pygfx.WorldObject(*args, **kwargs)
        self._container.add(self._fill, self._outline, self._handles)
        if parent:
            parent.add(self._container)

        # Utilities for moving ROI
        self._selected = False
        self._move_mode: ROIMoveMode | None = None
        # NB _move_anchor has different meanings depending on _move_mode
        self._move_anchor: tuple[float, float] = (0, 0)
        self._render: Callable = render
        self._canvas_to_world: Callable = canvas_to_world

        # Initialize
        self.set_fill(_cmap.Color("transparent"))
        self.set_border(_cmap.Color("yellow"))
        self.set_handles(_cmap.Color("white"))
        self.set_visible(False)

    # -- BoundingBox methods -- #

    def set_bounding_box(
        self, minimum: tuple[float, float], maximum: tuple[float, float]
    ) -> None:
        # NB: Support two diagonal points, not necessarily true min/max
        x1 = float(min(minimum[0], maximum[0]))
        y1 = float(min(minimum[1], maximum[1]))
        x2 = float(max(minimum[0], maximum[0]))
        y2 = float(max(minimum[1], maximum[1]))

        # Update each handle
        self._positions[0, :2] = [x1, y1]
        self._positions[1, :2] = [x2, y1]
        self._positions[2, :2] = [x2, y2]
        self._positions[3, :2] = [x1, y2]
        self._positions[4, :2] = [x1, y1]
        self._refresh()

    def set_fill(self, color: _cmap.Color) -> None:
        if self._fill:
            self._fill.material.color = color.rgba
            self._render()

    def set_border(self, color: _cmap.Color) -> None:
        if self._outline:
            self._outline.material.color = color.rgba
            self._render()

    # TODO: Misleading name?
    def set_handles(self, color: _cmap.Color) -> None:
        if self._handles:
            self._handles.material.color = color.rgba
            self._render()

    def _create_fill(self) -> pygfx.Mesh:
        fill = pygfx.Mesh(
            geometry=pygfx.Geometry(
                positions=self._positions,
                indices=np.array([[0, 1, 2, 3]], dtype=np.int32),
            ),
            material=pygfx.MeshBasicMaterial(),
        )
        return fill

    def _create_outline(self) -> pygfx.Line:
        outline = pygfx.Line(
            geometry=pygfx.Geometry(
                positions=self._positions,
                indices=np.array([[0, 1, 2, 3]], dtype=np.int32),
            ),
            material=pygfx.LineMaterial(thickness=1),
        )
        return outline

    def _create_handles(self) -> pygfx.Points:
        geometry = pygfx.Geometry(positions=self._positions[:-1])
        handles = pygfx.Points(
            geometry=geometry,
            # FIXME Size in pixels is not ideal for selection.
            # TODO investigate what size_mode = vertex does...
            material=pygfx.PointsMaterial(size=1.5 * self._handle_rad),  # pyright: ignore[reportArgumentType]
        )

        # NB: Default bounding box for points does not consider the radius of
        # those points. We need to HACK it for handle selection
        def get_handle_bb(
            old: Callable[[], np.ndarray | None],
        ) -> Callable[[], np.ndarray | None]:
            def new_get_bb() -> np.ndarray | None:
                if (bb := old()) is None:
                    return None
                bb = bb.copy()
                bb[0, :2] -= self._handle_rad
                bb[1, :2] += self._handle_rad
                return bb

            return new_get_bb

        handles.get_bounding_box = get_handle_bb(handles.get_bounding_box)
        return handles

    def can_select(self) -> bool:
        return True

    def selected(self) -> bool:
        return self._selected

    def set_selected(self, selected: bool) -> None:
        self._selected = selected
        if self._handles:
            self._handles.visible = selected

    def _refresh(self) -> None:
        if self._fill:
            self._fill.geometry.positions.data[:, :] = self._positions
            self._fill.geometry.positions.update_range()
        if self._outline:
            self._outline.geometry.positions.data[:, :] = self._positions
            self._outline.geometry.positions.update_range()
        if self._handles:
            self._handles.geometry.positions.data[:, :] = self._positions[:-1]
            self._handles.geometry.positions.update_range()
        self._render()

    def on_mouse_move(self, event: MouseMoveEvent) -> bool:
        # Convert canvas -> world
        world_pos = tuple(self._canvas_to_world((event.x, event.y))[:2])
        # moving a handle
        if self._move_mode == ROIMoveMode.HANDLE:
            # The anchor is set to the opposite handle, which never moves.
            self.boundingBoxChanged.emit((world_pos, self._move_anchor))
        # translating the whole roi
        elif self._move_mode == ROIMoveMode.TRANSLATE:
            # The anchor is the mouse position reported in the previous mouse event.
            dx = world_pos[0] - self._move_anchor[0]
            dy = world_pos[1] - self._move_anchor[1]
            # If the mouse moved (dx, dy) between events, the whole ROI needs to be
            # translated that amount.
            new_min = (self._positions[0, 0] + dx, self._positions[0, 1] + dy)
            new_max = (self._positions[2, 0] + dx, self._positions[2, 1] + dy)
            self.boundingBoxChanged.emit((new_min, new_max))
            self._move_anchor = world_pos

        return False

    def on_mouse_press(self, event: MousePressEvent) -> bool:
        self.set_selected(True)
        # Convert canvas -> world
        world_pos = self._canvas_to_world((event.x, event.y))
        drag_idx = self._handle_under(world_pos)
        # If a marker is pressed
        if drag_idx is not None:
            opposite_idx = (drag_idx + 2) % 4
            self._move_mode = ROIMoveMode.HANDLE
            self._move_anchor = tuple(self._positions[opposite_idx, :2].copy())
        # If the rectangle is pressed
        else:
            self._move_mode = ROIMoveMode.TRANSLATE
            self._move_anchor = world_pos
        return False

    def on_mouse_release(self, event: MouseReleaseEvent) -> bool:
        return False

    def visible(self) -> bool:
        if self._outline:
            return bool(self._outline.visible)
        if self._fill:
            return bool(self._fill.visible)
        # Nothing to see
        return False

    def set_visible(self, visible: bool) -> None:
        if fill := getattr(self, "_fill", None):
            fill.visible = visible
        if outline := getattr(self, "_outline", None):
            outline.visible = visible
        if handles := getattr(self, "_handles", None):
            handles.visible = visible and self.selected()
        self._render()

    def _handle_under(self, pos: Sequence[float]) -> int | None:
        """Returns an int in [0, 3], or None.

        If an int i, means that the handle at self._positions[i] is at pos.
        If None, there is no handle at pos.
        """
        # FIXME: Ideally, Renderer.get_pick_info would do this for us. But it
        # seems broken.
        for i, p in enumerate(self._positions[:-1]):
            if (p[0] - pos[0]) ** 2 + (p[1] - pos[1]) ** 2 <= self._handle_rad**2:
                return i
        return None

    def get_cursor(self, mme: MouseMoveEvent) -> CursorType | None:
        # Convert event pos (on canvas) to world pos
        world_pos = self._canvas_to_world((mme.x, mme.y))
        # Step 1: Handles
        # Preferred over the rectangle
        # Can only be moved if ROI is selected
        if (idx := self._handle_under(world_pos)) is not None and self.selected():
            # Idx 0 is top left, 2 is bottom right
            if idx % 2 == 0:
                return CursorType.FDIAG_ARROW
            # Idx 1 is bottom left, 3 is top right
            return CursorType.BDIAG_ARROW
        # Step 2: Entire ROI
        if self._outline:
            roi_bb = self._outline.get_bounding_box()
            if _is_inside(roi_bb, world_pos):
                return CursorType.ALL_ARROW
        return None

    def remove(self) -> None:
        if (par := self._container.parent) is not None:
            par.remove(self._container)


class GfxArrayCanvas(ArrayCanvas):
    """pygfx-based canvas wrapper."""

    def __init__(self, viewer_model: ArrayViewerModel) -> None:
        self._viewer = viewer_model

        self._current_shape: tuple[int, ...] = ()
        self._last_state: dict[Literal[2, 3], Any] = {}

        cls = rendercanvas_class()
        self._canvas = cls(size=(600, 600))

        # this filter needs to remain in scope for the lifetime of the canvas
        # or mouse events will not be intercepted
        # the returned function can be called to remove the filter, (and it also
        # closes on the event filter and keeps it in scope).
        self._disconnect_mouse_events = filter_mouse_events(self._canvas, self)

        self._renderer = pygfx.renderers.WgpuRenderer(self._canvas)

        self._scene = pygfx.Scene()
        self._scene.add(pygfx.Background(None, pygfx.BackgroundMaterial("black")))
        self._camera: pygfx.Camera | None = None
        self._ndim: Literal[2, 3] | None = None

        self._elements = WeakKeyDictionary[pygfx.WorldObject, CanvasElement]()
        self._selection: CanvasElement | None = None
        # Maintain a weak reference to the last ROI created.
        self._last_roi_created: ReferenceType[PyGFXRectangle] | None = None

        self._canvas.add_event_handler(lambda e: self.refresh(), "resize")

    def frontend_widget(self) -> Any:
        return self._canvas

    def set_ndim(self, ndim: Literal[2, 3]) -> None:
        """Set the number of dimensions of the displayed data."""
        if ndim == self._ndim:
            return
        elif self._ndim is not None and self._camera is not None:
            # remember the current state before switching to the new camera
            self._last_state[self._ndim] = self._camera.get_state()

        self._ndim = ndim
        if ndim == 3:
            self._camera = cam = pygfx.PerspectiveCamera(0, 1)
            with suppress(ValueError):
                # if the scene has no children yet, this will raise a ValueErrors
                # FIXME: there's a bit of order-of-call problem here:
                # this method needs to be called *after* the scene is constructed...
                # that's what controller._on_model_visible_axes_changed does, but
                # it seems fragile and should be fixed.
                cam.show_object(self._scene, up=(0, -1, 0), view_dir=(0, 0, 1))
            controller = pygfx.OrbitController(cam, register_events=self._renderer)
            zoom = "zoom"
            # FIXME: there is still an issue with rotational centration.
            # the controller is not rotating around the middle of the volume...
            # but I think it might actually be a pygfx issue... the critical state
            # seems to be somewhere outside of the camera's get_state dict.
        else:
            self._camera = cam = pygfx.OrthographicCamera(512, 512)
            cam.local.scale_y = -1
            cam.local.position = (256, 256, 0)
            controller = pygfx.PanZoomController(cam, register_events=self._renderer)
            zoom = "zoom_to_point"

        self._controller = controller
        # increase zoom wheel gain
        self._controller.controls.update({"wheel": (zoom, "push", -0.005)})

        # restore the previous state if it exists
        if state := self._last_state.get(ndim):
            cam.set_state(state)

    def add_image(self, data: np.ndarray | None = None) -> PyGFXImageHandle:
        """Add a new Image node to the scene."""
        if data is not None:
            # pygfx uses a view of the data without copy, so if we don't
            # copy it here, the original data will be modified when the
            # texture changes.
            data = data.copy()
        tex = pygfx.Texture(data, dim=2)
        image = pygfx.Image(
            pygfx.Geometry(grid=tex),
            pygfx.ImageBasicMaterial(depth_test=False, alpha_mode="add"),
        )
        self._scene.add(image)

        if data is not None:
            self._current_shape, prev_shape = data.shape, self._current_shape
            if not prev_shape:
                self.set_range()

        # FIXME: I suspect there are more performant ways to refresh the canvas
        # look into it.
        handle = PyGFXImageHandle(image, self.refresh)
        self._elements[image] = handle
        return handle

    def add_volume(self, data: np.ndarray | None = None) -> PyGFXImageHandle:
        if data is not None:
            # pygfx uses a view of the data without copy, so if we don't
            # copy it here, the original data will be modified when the
            # texture changes.
            data = data.copy()
        tex = pygfx.Texture(data, dim=3)
        vol = pygfx.Volume(
            pygfx.Geometry(grid=tex),
            pygfx.VolumeRayMaterial(
                interpolation="nearest", depth_test=False, alpha_mode="add"
            ),
        )
        self._scene.add(vol)

        if data is not None:
            vol.local_position = [-0.5 * i for i in data.shape[::-1]]
            self._current_shape, prev_shape = data.shape, self._current_shape
            if len(prev_shape) != 3:
                self.set_range()

        # FIXME: I suspect there are more performant ways to refresh the canvas
        # look into it.
        handle = PyGFXImageHandle(vol, self.refresh)
        self._elements[vol] = handle
        return handle

    def add_bounding_box(self) -> PyGFXRectangle:
        """Add a new Rectangular ROI node to the scene."""
        roi = PyGFXRectangle(
            render=self.refresh,
            canvas_to_world=self.canvas_to_world,
            parent=self._scene,
        )
        roi.set_visible(False)
        self._elements[roi._container] = roi
        self._last_roi_created = ref(roi)
        return roi

    def set_scales(self, scales: tuple[float, ...]) -> None:
        """Set per-visible-axis scale factors for rendering."""
        if not scales:
            return
        # scales are in data order (slowest-to-fastest, e.g. ZYX)
        # pygfx uses XYZ, so reverse
        gfx_scales = list(reversed(scales))
        # pad to 3 components
        while len(gfx_scales) < 3:
            gfx_scales.append(1.0)
        sx, sy, sz = gfx_scales[0], gfx_scales[1], gfx_scales[2]
        has_visuals = False
        for child in self._scene.children:
            if isinstance(child, (pygfx.Image, pygfx.Volume)):
                child.local.scale = (sx, sy, sz)
                has_visuals = True
        if has_visuals:
            self.set_range()

    def set_range(
        self,
        x: tuple[float, float] | None = None,
        y: tuple[float, float] | None = None,
        z: tuple[float, float] | None = None,
        margin: float = 0.05,
    ) -> None:
        """Update the range of the PanZoomCamera.

        When called with no arguments, the range is set to the full extent of the data.
        """
        if not self._scene.children or self._camera is None:
            return

        cam = self._camera
        cam.show_object(self._scene)

        if (bb := self._scene.get_world_bounding_box()) is not None:
            width, height, _depth = np.ptp(bb, axis=0)
            if width < 0.01:
                width = 1
            if height < 0.01:
                height = 1
            cam.width = width
            cam.height = height
            cam.zoom = 1 - margin
        self.refresh()

    def zoom(self, factor: float | tuple, center: tuple[float, float] = (0, 0)) -> None:
        """Zoom in (or out) at the given center (world coordinates)."""
        if (cam := self._camera) is None:
            return

        cx, cy = center
        px, py, pz = cam.local.position
        cam.local.position = (
            cx + (px - cx) * factor,
            cy + (py - cy) * factor,
            pz,
        )
        cam.zoom /= factor
        self._canvas.force_draw()

    def refresh(self) -> None:
        with suppress(AttributeError):
            self._canvas.update()
        self._canvas.request_draw(self._animate)

    def _animate(self) -> None:
        if self._camera is not None:
            self._renderer.render(self._scene, self._camera)

    def canvas_to_world(
        self, pos_xy: tuple[float, float]
    ) -> tuple[float, float, float]:
        """Map XY canvas position (pixels) to XYZ coordinate in world space."""
        # Code adapted from:
        # https://github.com/pygfx/pygfx/pull/753/files#diff-173d643434d575e67f8c0a5bf2d7ea9791e6e03a4e7a64aa5fa2cf4172af05cdR395
        viewport = pygfx.Viewport.from_viewport_or_renderer(self._renderer)
        if not viewport.is_inside(*pos_xy):
            return (-1, -1, -1)

        # Get position relative to viewport
        pos_rel = (
            pos_xy[0] - viewport.rect[0],
            pos_xy[1] - viewport.rect[1],
        )

        vs = viewport.logical_size

        # Convert position to NDC
        x = pos_rel[0] / vs[0] * 2 - 1
        y = -(pos_rel[1] / vs[1] * 2 - 1)
        pos_ndc = (x, y, 0)

        if self._camera:
            pos_ndc += la.vec_transform(
                self._camera.world.position, self._camera.camera_matrix
            )
            pos_world = la.vec_unproject(pos_ndc[:2], self._camera.camera_matrix)

            # NB In vispy, (0.5,0.5) is a center of an image pixel, while in pygfx
            # (0,0) is the center. We conform to vispy's standard.
            return (pos_world[0] + 0.5, pos_world[1] + 0.5, pos_world[2] + 0.5)
        else:
            return (-1, -1, -1)

    def elements_at(self, pos_xy: tuple[float, float]) -> list[CanvasElement]:
        """Obtains all elements located at pos."""
        # FIXME: Ideally, Renderer.get_pick_info would do this and
        # canvas_to_world for us. But it seems broken.
        elements: list[CanvasElement] = []
        pos = self.canvas_to_world((pos_xy[0], pos_xy[1]))
        for c in self._scene.children:
            bb = c.get_bounding_box()
            if _is_inside(bb, pos):
                elements.append(self._elements[c])
        return elements

    def set_visible(self, visible: bool) -> None:
        """Set the visibility of the canvas."""
        self._canvas.visible = visible

    def close(self) -> None:
        self._disconnect_mouse_events()
        self._canvas.close()

    def on_mouse_press(self, event: MousePressEvent) -> bool:
        if self._selection:
            self._selection.set_selected(False)
            self._selection = None
        canvas_pos = (event.x, event.y)
        world_pos = self.canvas_to_world(canvas_pos)[:2]

        # If in CREATE_ROI mode, the new ROI should "start" here.
        if self._viewer.interaction_mode == InteractionMode.CREATE_ROI:
            if self._last_roi_created is None:
                raise ValueError("No ROI to create!")
            if new_roi := self._last_roi_created():
                self._last_roi_created = None
                # HACK: Provide a non-zero starting size so that if the user clicks
                # and immediately releases, it's visible and can be selected again
                _min = world_pos
                _max = (world_pos[0] + 1, world_pos[1] + 1)
                # Put the ROI where the user clicked
                new_roi.boundingBoxChanged.emit((_min, _max))
                # Make it visible
                new_roi.set_visible(True)
                # Select it so the mouse press event below triggers ROIMoveMode.HANDLE
                # TODO: Make behavior more direct
                new_roi.set_selected(True)

            # All done - exit the mode
            self._viewer.interaction_mode = InteractionMode.PAN_ZOOM

        # Select first selectable object at clicked point
        for vis in self.elements_at(canvas_pos):
            if vis.can_select():
                self._selection = vis
                self._selection.on_mouse_press(event)
                return False

        return False

    def on_mouse_move(self, event: MouseMoveEvent) -> bool:
        if event.btn == MouseButton.LEFT:
            if self._selection and self._selection.selected():
                self._selection.on_mouse_move(event)
                # If we are moving the object, we don't want to move the camera
                return True
        return False

    def on_mouse_release(self, event: MouseReleaseEvent) -> bool:
        if self._selection:
            self._selection.on_mouse_release(event)
        return False

    def get_cursor(self, event: MouseMoveEvent) -> CursorType:
        if self._viewer.interaction_mode == InteractionMode.CREATE_ROI:
            return CursorType.CROSS
        for vis in self.elements_at((event.x, event.y)):
            if cursor := vis.get_cursor(event):
                return cursor
        return CursorType.DEFAULT



from __future__ import annotations

import numpy as np
import pytest

from ndv.models._viewer_model import ArrayViewerModel
from ndv.views._pygfx._array_canvas import GfxArrayCanvas


def _force_canvas_size(canvas: GfxArrayCanvas, w: int = 600, h: int = 600) -> None:
    """Force the rendercanvas to report a valid size (needed before show)."""
    rc = canvas._canvas
    rc._size_info.set_physical_size(w, h, 1.0)


@pytest.mark.usefixtures("any_app")
def test_zoom_center() -> None:
    """Zoom should keep the center point fixed in world space."""
    canvas = GfxArrayCanvas(ArrayViewerModel())
    _force_canvas_size(canvas)
    canvas.set_ndim(2)
    canvas.add_image(np.random.rand(100, 100).astype(np.float32))
    canvas.set_range()

    cam = canvas._camera
    assert cam is not None

    # Record initial state (copy position since it's mutable)
    initial_pos = tuple(cam.local.position)
    initial_zoom = cam.zoom

    # Zoom in at a specific world point
    center = (30.0, 70.0)
    canvas.zoom(factor=0.5, center=center)

    # Camera zoom should have increased (zoomed in)
    assert cam.zoom > initial_zoom

    # Camera should have moved toward the center point
    new_pos = tuple(cam.local.position)
    assert abs(center[0] - new_pos[0]) < abs(center[0] - initial_pos[0])
    assert abs(center[1] - new_pos[1]) < abs(center[1] - initial_pos[1])

    # Zoom back out by the inverse factor
    canvas.zoom(factor=2.0, center=center)

    # Should return approximately to initial state
    restored_pos = tuple(cam.local.position)
    assert restored_pos[:2] == pytest.approx(initial_pos[:2])
    assert cam.zoom == pytest.approx(initial_zoom)

    canvas.close()



import numpy as np
import pytest

from ndv.models._ring_buffer import RingBuffer


def test_dtype() -> None:
    r = RingBuffer(5)
    assert r.dtype == np.dtype(np.float64)

    r = RingBuffer(5, dtype=bool)
    assert r.dtype == np.dtype(bool)


def test_sizes() -> None:
    rb = RingBuffer(5, dtype=(int, 2))
    assert rb.maxlen == 5
    assert len(rb) == 0
    assert rb.shape == (0, 2)

    rb.append([0, 0])
    assert rb.maxlen == 5
    assert len(rb) == 1
    assert rb.shape == (1, 2)


def test_append() -> None:
    rb = RingBuffer(5)

    rb.append(1)
    np.testing.assert_equal(rb, np.array([1]))
    assert len(rb) == 1

    rb.append(2)
    np.testing.assert_equal(rb, np.array([1, 2]))
    assert len(rb) == 2

    rb.append(3)
    rb.append(4)
    rb.append(5)
    np.testing.assert_equal(rb, np.array([1, 2, 3, 4, 5]))
    assert len(rb) == 5

    rb.append(6)
    np.testing.assert_equal(rb, np.array([2, 3, 4, 5, 6]))
    assert len(rb) == 5

    assert rb[4] == 6
    assert rb[-1] == 6


def test_getitem() -> None:
    rb = RingBuffer(5)
    rb.extend([1, 2, 3])
    rb.extendleft([4, 5])
    expected = np.array([4, 5, 1, 2, 3])
    np.testing.assert_equal(rb, expected)

    for i in range(rb.maxlen):
        assert expected[i] == rb[i]

    ii = [0, 4, 3, 1, 2]
    np.testing.assert_equal(rb[ii], expected[ii])


def test_getitem_negative_index() -> None:
    rb = RingBuffer(5)
    rb.extend([1, 2, 3])
    assert rb[-1] == 3


def test_appendleft() -> None:
    rb = RingBuffer(5)

    rb.appendleft(1)
    np.testing.assert_equal(rb, np.array([1]))
    assert len(rb) == 1

    rb.appendleft(2)
    np.testing.assert_equal(rb, np.array([2, 1]))
    assert len(rb) == 2

    rb.appendleft(3)
    rb.appendleft(4)
    rb.appendleft(5)
    np.testing.assert_equal(rb, np.array([5, 4, 3, 2, 1]))
    assert len(rb) == 5

    rb.appendleft(6)
    np.testing.assert_equal(rb, np.array([6, 5, 4, 3, 2]))
    assert len(rb) == 5


def test_extend() -> None:
    rb = RingBuffer(5)
    rb.extend([1, 2, 3])
    np.testing.assert_equal(rb, np.array([1, 2, 3]))
    rb.popleft()
    rb.extend([4, 5, 6])
    np.testing.assert_equal(rb, np.array([2, 3, 4, 5, 6]))
    rb.extendleft([0, 1])
    np.testing.assert_equal(rb, np.array([0, 1, 2, 3, 4]))

    rb.extendleft([1, 2, 3, 4, 5, 6, 7])
    np.testing.assert_equal(rb, np.array([1, 2, 3, 4, 5]))

    rb.extend([1, 2, 3, 4, 5, 6, 7])
    np.testing.assert_equal(rb, np.array([3, 4, 5, 6, 7]))


def test_pops() -> None:
    rb = RingBuffer(3)
    rb.append(1)
    rb.appendleft(2)
    rb.append(3)
    np.testing.assert_equal(rb, np.array([2, 1, 3]))

    assert rb.pop() == 3
    np.testing.assert_equal(rb, np.array([2, 1]))

    assert rb.popleft() == 2
    np.testing.assert_equal(rb, np.array([1]))

    # test empty pops
    empty = RingBuffer(1)
    with pytest.raises(IndexError, match="pop from an empty RingBuffer"):
        empty.pop()
    with pytest.raises(IndexError, match="pop from an empty RingBuffer"):
        empty.popleft()


def test_2d() -> None:
    rb = RingBuffer(5, dtype=(float, 2))

    rb.append([1, 2])
    np.testing.assert_equal(rb, np.array([[1, 2]]))
    assert len(rb) == 1
    assert np.shape(rb) == (1, 2)

    rb.append([3, 4])
    np.testing.assert_equal(rb, np.array([[1, 2], [3, 4]]))
    assert len(rb) == 2
    assert np.shape(rb) == (2, 2)

    rb.appendleft([5, 6])
    np.testing.assert_equal(rb, np.array([[5, 6], [1, 2], [3, 4]]))
    assert len(rb) == 3
    assert np.shape(rb) == (3, 2)

    np.testing.assert_equal(rb[0], [5, 6])
    np.testing.assert_equal(rb[0, :], [5, 6])
    np.testing.assert_equal(rb[:, 0], [5, 1, 3])


def test_3d() -> None:
    np.random.seed(0)
    frame_shape = (32, 32)

    rb = RingBuffer(5, dtype=("u2", frame_shape))
    frame = np.random.randint(0, 65535, frame_shape, dtype="u2")
    rb.append(frame)
    np.testing.assert_equal(rb, frame[None])
    frame2 = np.random.randint(0, 65535, frame_shape, dtype="u2")
    rb.append(frame2)
    np.testing.assert_equal(rb[-1], frame2)
    np.testing.assert_equal(rb, np.array([frame, frame2]))

    # fill buffer
    for _ in range(5):
        rb.append(np.random.randint(0, 65535, frame_shape, dtype="u2"))

    # add known frame
    frame3 = np.random.randint(0, 65535, frame_shape, dtype="u2")
    rb.append(frame3)
    np.testing.assert_equal(rb[-1], frame3)


def test_iter() -> None:
    rb = RingBuffer(5)
    for i in range(5):
        rb.append(i)
    for i, j in zip(rb, range(5), strict=False):
        assert i == j


def test_repr() -> None:
    rb = RingBuffer(5, dtype=int)
    for i in range(5):
        rb.append(i)

    assert repr(rb) == "<RingBuffer of array([0, 1, 2, 3, 4])>"


def test_no_overwrite() -> None:
    rb = RingBuffer(3, allow_overwrite=False)
    rb.append(1)
    rb.append(2)
    rb.appendleft(3)
    with pytest.raises(IndexError, match="overwrite"):
        rb.appendleft(4)
    with pytest.raises(IndexError, match="overwrite"):
        rb.extendleft([4])
    rb.extendleft([])

    np.testing.assert_equal(rb, np.array([3, 1, 2]))
    with pytest.raises(IndexError, match="overwrite"):
        rb.append(4)
    with pytest.raises(IndexError, match="overwrite"):
        rb.extend([4])
    rb.extend([])

    # works fine if we pop the surplus
    rb.pop()
    rb.append(4)
    np.testing.assert_equal(rb, np.array([3, 1, 4]))


def test_degenerate() -> None:
    r = RingBuffer(0)
    np.testing.assert_equal(r, np.array([]))

    # this does not error with deque(maxlen=0), so should not error here
    try:
        r.append(0)
        r.appendleft(0)
    except IndexError:
        pytest.fail("IndexError raised when appending to a degenerate RingBuffer")



import numpy as np
from qtpy.QtWidgets import QApplication
from vispy import scene

qapp = QApplication([])
canvas = scene.SceneCanvas(keys="interactive", size=(800, 600))

# Set up a viewbox to display the image with interactive pan/zoom
view = canvas.central_widget.add_view()

# Create the image
img_data = np.random.rand(512, 512).astype(np.float32)

image = scene.visuals.Image(img_data, interpolation="nearest", parent=view.scene)
view.camera = scene.PanZoomCamera(aspect=1)
view.camera.flip = (0, 1, 0)
view.camera.set_range()

canvas.show()
qapp.exec()



"""General model for ndv."""

import warnings
from enum import Enum
from typing import TYPE_CHECKING, Any, Literal, TypeAlias, TypedDict, cast

from cmap import Colormap
from pydantic import Field, computed_field, model_validator
from typing_extensions import Self

from ndv._types import AxisKey, ChannelKey, Slice

from ._base_model import NDVModel
from ._lut_model import LUTModel
from ._mapping import ValidatedEventedDict
from ._reducer import ReducerType

if TYPE_CHECKING:
    from collections.abc import (  # noqa: F401 # used for mkdocstrings
        Callable,  # used for mkdocstrings
        Hashable,
        Mapping,
        Sequence,
    )

    import cmap
    import numpy.typing as npt  # noqa: F401  # used for mkdocstrings

    from ._lut_model import AutoscaleType

    class LUTModelKwargs(TypedDict, total=False):
        """Keyword arguments for `LUTModel`."""

        name: str
        visible: bool
        cmap: "cmap.Colormap | cmap._colormap.ColorStopsLike"
        clims: "tuple[float, float] | None"
        gamma: float
        autoscale: AutoscaleType

    class ArrayDisplayModelKwargs(TypedDict, total=False):
        """Keyword arguments for `ArrayDisplayModel`."""

        visible_axes: "tuple[AxisKey, AxisKey, AxisKey] | tuple[AxisKey, AxisKey]"
        current_index: Mapping[AxisKey, int | slice]
        channel_mode: 'ChannelMode | Literal["grayscale", "composite", "color", "rgba"]'
        channel_axis: AxisKey | None
        reducers: Mapping["AxisKey | None", ReducerType]
        luts: Mapping["int | None", "LUTModel | LUTModelKwargs"]
        default_lut: "LUTModel | LUTModelKwargs"
        scales: Mapping[AxisKey, float] | Sequence[float]


# map of axis to index/slice ... i.e. the current subset of data being displayed
IndexMap: TypeAlias = ValidatedEventedDict[AxisKey, int | Slice]
# map of index along channel axis to LUTModel object
LUTMap: TypeAlias = ValidatedEventedDict[ChannelKey, LUTModel]
# map of axis to reducer
Reducers: TypeAlias = ValidatedEventedDict[AxisKey | None, ReducerType]
# map of axis to scale factor
ScalesMap: TypeAlias = ValidatedEventedDict[AxisKey, float]
# used for visible_axes
TwoOrThreeAxisTuple: TypeAlias = (
    tuple[AxisKey, AxisKey, AxisKey] | tuple[AxisKey, AxisKey]
)


DEFAULT_LUT_COLORS = ("green", "magenta", "cyan", "red", "blue", "yellow")


def _default_luts() -> LUTMap:
    return ValidatedEventedDict(
        (i, LUTModel(cmap=Colormap(color)))
        for i, color in enumerate(DEFAULT_LUT_COLORS)
    )


class ChannelMode(str, Enum):
    """Channel display mode.

    Attributes
    ----------
    GRAYSCALE : str
        The array is displayed as a single channel, with a single lookup table applied.
        In this mode, there effective *is* no channel axis: all non-visible dimensions
        have sliders, and there is a single LUT control (the `default_lut`).
    COMPOSITE : str
        Display all (or a subset of) channels as a composite image, with a different
        lookup table applied to each channel.  In this mode, the slider for the channel
        axis is hidden by default, and LUT controls for each channel are shown.
    COLOR : str
        Display a single channel at a time as a color image, with a channel-specific
        lookup table applied.  In this mode, the slider for the channel axis is shown,
        and the user can select which channel to display.  LUT controls are shown for
        all channels.
    RGBA : str
        The array is displayed as an RGB image, with a single lookup table applied.
        In this mode, the slider for the channel axis is hidden, and a single LUT
        control is shown. Only valid when channel axis has length <= 4.
    RGB : str
        Alias for RGBA.
    """

    GRAYSCALE = "grayscale"
    COMPOSITE = "composite"
    COLOR = "color"
    RGBA = "rgba"

    def __str__(self) -> str:
        return self.value

    def is_multichannel(self) -> bool:
        """Return whether this mode displays multiple channels.

        If `is_multichannel` is True, then the `channel_axis` slider should be hidden.
        """
        return self in (ChannelMode.COMPOSITE, ChannelMode.RGBA)


ChannelMode._member_map_["RGB"] = ChannelMode.RGBA  #  ChannelMode["RGB"]
ChannelMode._value2member_map_["rgb"] = ChannelMode.RGBA  # ChannelMode("rgb")


class ArrayDisplayModel(NDVModel):
    """Model of how to display an array.

    An `ArrayDisplayModel` is used to specify how to display a multi-dimensional array.
    It specifies which axes are visible, how to reduce along axes that are not visible,
    how to display channels, and how to apply lookup tables to channels.  It is
    typically paired with a [`ndv.DataWrapper`][] in order to resolve axis keys and
    slice data.

    !!! info

        In the following types, `Hashable` is used to refer to a type that will
        typically be either an integer index or a string label for an axis.

    Attributes
    ----------
    visible_axes : tuple[Hashable, ...]
        Ordered list of axes to visualize, from slowest to fastest.
        e.g. `('Z', -2, -1)`
    current_index : Mapping[Hashable, int | slice]
        The currently displayed position/slice along each dimension.
        e.g. {0: 0, 'time': slice(10, 20)}
        Not all axes need be present, and axes not present are assumed to
        be slice(None), meaning it is up to the controller of this model to restrict
        indices to an efficient range for retrieval.
        If the number of non-singleton axes is greater than `n_visible_axes`,
        then `reducers` are used to reduce the data along the remaining axes.
    reducers : Mapping[Hashable | None, numpy.ufunc]
        Function used to reduce data along axes remaining after slicing.
        Ideally, the ufunc should accept an `axis` argument.
        *(TODO: what happens if not?)*
    default_reducer : numpy.ufunc
        Default reducer to use when no reducer is specified for an axis. By default,
        this is [`numpy.max`][].
    channel_mode : ChannelMode
        How to display channel information:

        - `GRAYSCALE`: ignore channel axis, use `default_lut`
        - `COMPOSITE`: display all channels as a composite image, using `luts`
        - `COLOR`: display a single channel at a time, using `luts`
        - `RGBA`: display as an RGB image, using `default_lut` (except for cmap)

        If `channel_mode` is set to anything other than `GRAYSCALE`, then `channel_axis`
        must be set to a valid axis; if no `channel_axis` is set, at the time of
        display, the [`DataWrapper`][ndv.DataWrapper] MAY guess the `channel_axis`,
        and set it on the model.
    channel_axis : Hashable | None
        The dimension index or name of the channel dimension.
        The implication of setting channel_axis is that *all* elements along the channel
        dimension are shown, with different LUTs applied to each channel.
        If None, then a single lookup table is used for all channels (`luts[None]`).
        NOTE: it is an error for channel_axis to be in `visible_axes` (or ignore it?)
    luts : Mapping[int | None, LUTModel]
        Instructions for how to display each channel of the array.
        Keys represent position along the dimension specified by `channel_axis`.
        Values are `LUT` objects that specify how to display the channel.
        The special key `None` is used to represent a fallback LUT for all channels,
        and is used when `channel_axis` is None.  It should always be present
    default_lut : LUTModel
        Default lookup table to use when no `LUTModel` is specified for a channel in
        `luts`.
    """

    # NOTE: In terms of requesting data, there is a slight "delocalization" of state
    # here in that we probably also want to avoid requesting data for channel
    # positions that are not visible.
    current_index: IndexMap = Field(default_factory=IndexMap, frozen=True)

    # map of axis to reducer (function that can reduce dimensionality along that axis)
    reducers: Reducers = Field(default_factory=Reducers, frozen=True)
    default_reducer: ReducerType = "numpy.max"  # type: ignore [assignment]  # FIXME

    channel_mode: ChannelMode = ChannelMode.GRAYSCALE
    channel_axis: AxisKey | None = None
    # must come after channel_axis, since it is used to set default visible_axes
    visible_axes: TwoOrThreeAxisTuple = Field(
        default_factory=lambda k: (-3, -2) if k.get("channel_axis") == -1 else (-2, -1)
    )

    # map of index along channel axis to LUTModel object
    luts: LUTMap = Field(default_factory=_default_luts)
    default_lut: LUTModel = Field(default_factory=LUTModel, frozen=True)

    # per-axis scale factors (e.g. physical pixel size)
    scales: ScalesMap = Field(default_factory=ScalesMap, frozen=True)

    @computed_field  # type: ignore [prop-decorator]
    @property
    def n_visible_axes(self) -> Literal[2, 3]:
        """Number of dims is derived from the length of `visible_axes`."""
        return cast("Literal[2, 3]", len(self.visible_axes))

    def model_post_init(self, context: Any, /) -> None:
        super().model_post_init(context)
        # apply default color cycle to luts without an explicit cmap
        # this is basically the same behavior as _default_luts,
        # but applies in the case where the user has provided partial lut
        # information (e.g. just the name) but not a cmap
        for key, lut in self.luts.items():
            if isinstance(key, int) and "cmap" not in lut.model_fields_set:
                color = DEFAULT_LUT_COLORS[key % len(DEFAULT_LUT_COLORS)]
                lut.cmap = Colormap(color)

    @model_validator(mode="after")
    def _validate_model(self) -> "Self":
        # prevent channel_axis from being in visible_axes
        if self.channel_axis in self.visible_axes:
            warnings.warn(
                f"Channel_axis {self.channel_axis!r} "
                f"cannot be in visible_axes {self.visible_axes!r}. "
                "Setting channel_axis to None.",
                UserWarning,
                stacklevel=2,
            )
            self.channel_axis = None

        # RGBA mode is not supported with 3D views
        if self.channel_mode == ChannelMode.RGBA and len(self.visible_axes) > 2:
            warnings.warn(
                "Cannot use RGBA mode with 3D view",
                stacklevel=2,
            )
            self.channel_mode = ChannelMode.GRAYSCALE

        return self
